# Copyright (C) 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from .. import mail_client, osutils, tests, urlutils


class TestMutt(tests.TestCase):
    def test_commandline(self):
        mutt = mail_client.Mutt(None)
        commandline = mutt._get_compose_commandline(None, None, "file%", body="hello")
        # The temporary filename is randomly generated, so it is not matched.
        self.assertEqual(["-a", "file%", "-i"], commandline[:-1])
        commandline = mutt._get_compose_commandline(
            "jrandom@example.org", "Hi there!", None
        )
        self.assertEqual(["-s", "Hi there!", "--", "jrandom@example.org"], commandline)

    def test_commandline_is_8bit(self):
        mutt = mail_client.Mutt(None)
        cmdline = mutt._get_compose_commandline(
            "jrandom@example.org", "Hi there!", "file%"
        )
        self.assertEqual(
            ["-s", "Hi there!", "-a", "file%", "--", "jrandom@example.org"], cmdline
        )
        for item in cmdline:
            self.assertTrue(
                isinstance(item, str),
                "Command-line item {!r} is not a native string!".format(item),
            )


class TestThunderbird(tests.TestCase):
    def test_commandline(self):
        tbird = mail_client.Thunderbird(None)
        commandline = tbird._get_compose_commandline(None, None, "file%")
        self.assertEqual(
            ["-compose", "attachment='{}'".format(urlutils.local_path_to_url("file%"))],
            commandline,
        )
        commandline = tbird._get_compose_commandline(
            "jrandom@example.org", "Hi there!", None, "bo'dy"
        )
        self.assertEqual(
            [
                "-compose",
                "body=bo%27dy,subject='Hi there!',to='jrandom@example.org'",
            ],
            commandline,
        )

    def test_commandline_is_8bit(self):
        # test for bug #139318
        tbird = mail_client.Thunderbird(None)
        cmdline = tbird._get_compose_commandline(
            "jrandom@example.org", "Hi there!", "file%"
        )
        self.assertEqual(
            [
                "-compose",
                ("attachment='{}',".format(urlutils.local_path_to_url("file%")))
                + "subject='Hi there!',to='jrandom@example.org'",
            ],
            cmdline,
        )
        for item in cmdline:
            self.assertTrue(
                isinstance(item, str),
                "Command-line item {!r} is not a native string!".format(item),
            )


class TestEmacsMail(tests.TestCase):
    def test_commandline(self):
        eclient = mail_client.EmacsMail(None)

        commandline = eclient._get_compose_commandline(None, "Hi there!", None)
        self.assertEqual(["--eval", '(compose-mail nil "Hi there!")'], commandline)

        commandline = eclient._get_compose_commandline(
            "jrandom@example.org", "Hi there!", None
        )
        self.assertEqual(
            ["--eval", '(compose-mail "jrandom@example.org" "Hi there!")'], commandline
        )

        # We won't be able to know the temporary file name at this stage
        # so we can't raise an assertion with assertEqual
        cmdline = eclient._get_compose_commandline(None, None, "file%")
        if eclient.elisp_tmp_file is not None:
            self.addCleanup(osutils.delete_any, eclient.elisp_tmp_file)
        commandline = " ".join(cmdline)
        self.assertContainsRe(commandline, "--eval")
        self.assertContainsRe(commandline, "(compose-mail nil nil)")
        self.assertContainsRe(commandline, "(load .*)")
        self.assertContainsRe(commandline, '(bzr-add-mime-att "file%")')

    def test_commandline_is_8bit(self):
        eclient = mail_client.EmacsMail(None)
        commandline = eclient._get_compose_commandline(
            "jrandom@example.org", "Hi there!", "file%"
        )
        if eclient.elisp_tmp_file is not None:
            self.addCleanup(osutils.delete_any, eclient.elisp_tmp_file)
        for item in commandline:
            self.assertTrue(
                isinstance(item, str),
                "Command-line item {!r} is not a native string!".format(item),
            )


class TestXDGEmail(tests.TestCase):
    def test_commandline(self):
        xdg_email = mail_client.XDGEmail(None)
        self.assertRaises(
            mail_client.NoMailAddressSpecified,
            xdg_email._get_compose_commandline,
            None,
            None,
            "file%",
        )
        commandline = xdg_email._get_compose_commandline(
            "jrandom@example.org", None, "file%"
        )
        self.assertEqual(["jrandom@example.org", "--attach", "file%"], commandline)
        commandline = xdg_email._get_compose_commandline(
            "jrandom@example.org", "Hi there!", None, "bo'dy"
        )
        self.assertEqual(
            ["jrandom@example.org", "--subject", "Hi there!", "--body", "bo'dy"],
            commandline,
        )

    def test_commandline_is_8bit(self):
        xdg_email = mail_client.XDGEmail(None)
        cmdline = xdg_email._get_compose_commandline(
            "jrandom@example.org", "Hi there!", "file%"
        )
        self.assertEqual(
            ["jrandom@example.org", "--subject", "Hi there!", "--attach", "file%"],
            cmdline,
        )
        for item in cmdline:
            self.assertTrue(
                isinstance(item, str),
                "Command-line item {!r} is not a native string!".format(item),
            )


class TestEvolution(tests.TestCase):
    def test_commandline(self):
        evo = mail_client.Evolution(None)
        commandline = evo._get_compose_commandline(None, None, "file%")
        self.assertEqual(["mailto:?attach=file%25"], commandline)
        commandline = evo._get_compose_commandline(
            "jrandom@example.org", "Hi there!", None, "bo&dy"
        )
        self.assertEqual(
            ["mailto:jrandom@example.org?body=bo%26dy&subject=Hi%20there%21"],
            commandline,
        )

    def test_commandline_is_8bit(self):
        evo = mail_client.Evolution(None)
        cmdline = evo._get_compose_commandline(
            "jrandom@example.org", "Hi there!", "file%"
        )
        self.assertEqual(
            ["mailto:jrandom@example.org?attach=file%25&subject=Hi%20there%21"], cmdline
        )
        for item in cmdline:
            self.assertTrue(
                isinstance(item, str),
                "Command-line item {!r} is not a native string!".format(item),
            )


class TestKMail(tests.TestCase):
    def test_commandline(self):
        kmail = mail_client.KMail(None)
        commandline = kmail._get_compose_commandline(None, None, "file%")
        self.assertEqual(["--attach", "file%"], commandline)
        commandline = kmail._get_compose_commandline(
            "jrandom@example.org", "Hi there!", None
        )
        self.assertEqual(["-s", "Hi there!", "jrandom@example.org"], commandline)

    def test_commandline_is_8bit(self):
        kmail = mail_client.KMail(None)
        cmdline = kmail._get_compose_commandline(
            "jrandom@example.org", "Hi there!", "file%"
        )
        self.assertEqual(
            ["-s", "Hi there!", "--attach", "file%", "jrandom@example.org"], cmdline
        )
        for item in cmdline:
            self.assertTrue(
                isinstance(item, str),
                "Command-line item {!r} is not a native string!".format(item),
            )


class TestClaws(tests.TestCase):
    def test_commandline(self):
        claws = mail_client.Claws(None)
        commandline = claws._get_compose_commandline(
            "jrandom@example.org", None, "file%"
        )
        self.assertEqual(
            ["--compose", "mailto:jrandom@example.org?", "--attach", "file%"],
            commandline,
        )
        commandline = claws._get_compose_commandline(
            "jrandom@example.org", "Hi there!", None
        )
        self.assertEqual(
            ["--compose", "mailto:jrandom@example.org?subject=Hi%20there%21"],
            commandline,
        )

    def test_commandline_is_8bit(self):
        claws = mail_client.Claws(None)
        cmdline = claws._get_compose_commandline(
            "jrandom@example.org", "\xb5cosm of fun!", "file%"
        )
        subject_string = urlutils.quote(
            "\xb5cosm of fun!".encode(osutils.get_user_encoding(), "replace")
        )
        self.assertEqual(
            [
                "--compose",
                "mailto:jrandom@example.org?subject={}".format(subject_string),
                "--attach",
                "file%",
            ],
            cmdline,
        )
        for item in cmdline:
            self.assertTrue(
                isinstance(item, str),
                "Command-line item {!r} is not a native string!".format(item),
            )

    def test_with_from(self):
        claws = mail_client.Claws(None)
        cmdline = claws._get_compose_commandline(
            "jrandom@example.org", None, None, None, "qrandom@example.com"
        )
        self.assertEqual(
            ["--compose", "mailto:jrandom@example.org?from=qrandom%40example.com"],
            cmdline,
        )

    def test_to_required(self):
        claws = mail_client.Claws(None)
        self.assertRaises(
            mail_client.NoMailAddressSpecified,
            claws._get_compose_commandline,
            None,
            None,
            "file%",
        )

    def test_with_body(self):
        claws = mail_client.Claws(None)
        cmdline = claws._get_compose_commandline(
            "jrandom@example.org", None, None, "This is some body text"
        )
        self.assertEqual(
            [
                "--compose",
                "mailto:jrandom@example.org?body=This%20is%20some%20body%20text",
            ],
            cmdline,
        )


class TestEditor(tests.TestCase):
    def test_get_merge_prompt_unicode(self):
        """Prompt, to and subject are unicode, the attachement is binary."""
        editor = mail_client.Editor(None)
        prompt = editor._get_merge_prompt(
            "foo\u1234", "bar\u1234", "baz\u1234", "qux\u1234".encode()
        )
        self.assertContainsRe(
            prompt, "foo\u1234(.|\n)*bar\u1234(.|\n)*baz\u1234(.|\n)*qux\u1234"
        )
        editor._get_merge_prompt("foo", "bar", "baz", b"qux\xff")


class DummyMailClient:
    def compose_merge_request(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs


class DefaultMailDummyClient(mail_client.DefaultMail):
    def __init__(self):
        self.client = DummyMailClient()

    def _mail_client(self):
        return self.client


class TestDefaultMail(tests.TestCase):
    def test_compose_merge_request(self):
        client = DefaultMailDummyClient()
        to = "a@b.com"
        subject = "[MERGE]"
        directive = ("directive",)
        basename = "merge"
        client.compose_merge_request(to, subject, directive, basename=basename)
        dummy_client = client.client
        self.assertEqual(dummy_client.args, (to, subject, directive))
        self.assertEqual(dummy_client.kwargs, {"basename": basename, "body": None})



# Copyright (C) 2008-2012 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""UI helper for the push command."""

from . import branch as _mod_branch
from . import controldir, errors, transport
from . import revision as _mod_revision
from .i18n import gettext
from .trace import note, warning


class PushResult:
    """Result of a push operation.

    :ivar branch_push_result: Result of a push between branches
    :ivar target_branch: The target branch
    :ivar stacked_on: URL of the branch on which the result is stacked
    :ivar workingtree_updated: Whether or not the target workingtree was updated.
    """

    def __init__(self):
        self.branch_push_result = None
        self.stacked_on = None
        self.workingtree_updated = None
        self.target_branch = None

    def report(self, to_file):
        """Write a human-readable description of the result."""
        if self.branch_push_result is None:
            if self.stacked_on is not None:
                note(
                    gettext("Created new stacked branch referring to %s.")
                    % self.stacked_on
                )
            else:
                note(gettext("Created new branch."))
        else:
            self.branch_push_result.report(to_file)


def _show_push_branch(
    br_from,
    revision_id,
    location,
    to_file,
    verbose=False,
    overwrite=False,
    remember=False,
    stacked_on=None,
    create_prefix=False,
    use_existing_dir=False,
    no_tree=False,
    lossy=False,
):
    """Push a branch to a location.

    :param br_from: the source branch
    :param revision_id: the revision-id to push up to
    :param location: the url of the destination
    :param to_file: the output stream
    :param verbose: if True, display more output than normal
    :param overwrite: list of things to overwrite ("history", "tags")
        or boolean indicating for everything
    :param remember: if True, store the location as the push location for
        the source branch
    :param stacked_on: the url of the branch, if any, to stack on;
        if set, only the revisions not in that branch are pushed
    :param create_prefix: if True, create the necessary parent directories
        at the destination if they don't already exist
    :param use_existing_dir: if True, proceed even if the destination
        directory exists without a current control directory in it
    :param lossy: Allow lossy push
    """
    to_transport = transport.get_transport(location, purpose="write")
    try:
        dir_to = controldir.ControlDir.open_from_transport(to_transport)
    except errors.NotBranchError:
        # Didn't find anything
        dir_to = None

    if dir_to is None:
        try:
            br_to = br_from.create_clone_on_transport(
                to_transport,
                revision_id=revision_id,
                stacked_on=stacked_on,
                create_prefix=create_prefix,
                use_existing_dir=use_existing_dir,
                no_tree=no_tree,
            )
        except errors.AlreadyControlDirError:
            raise errors.CommandError(
                gettext(
                    "Target directory %s already contains a .bzr directory, "
                    "but it is not valid."
                )
                % (location,)
            )
        except transport.FileExists:
            if not use_existing_dir:
                raise errors.CommandError(
                    gettext(
                        "Target directory %s"
                        " already exists, but does not have a .bzr"
                        " directory. Supply --use-existing-dir to push"
                        " there anyway."
                    )
                    % location
                )
            # This shouldn't occur, but if it does the FileExists error will be
            # more informative than an UnboundLocalError for br_to.
            raise
        except transport.NoSuchFile:
            if not create_prefix:
                raise errors.CommandError(
                    gettext(
                        "Parent directory of %s"
                        " does not exist."
                        "\nYou may supply --create-prefix to create all"
                        " leading parent directories."
                    )
                    % location
                )
            # This shouldn't occur (because create_prefix is true, so
            # create_clone_on_transport should be catching NoSuchFile and
            # creating the missing directories) but if it does the original
            # NoSuchFile error will be more informative than an
            # UnboundLocalError for br_to.
            raise
        except errors.TooManyRedirections:
            raise errors.CommandError(
                gettext("Too many redirections trying to make %s.") % location
            )
        push_result = PushResult()
        # TODO: Some more useful message about what was copied
        try:
            push_result.stacked_on = br_to.get_stacked_on_url()
        except (
            _mod_branch.UnstackableBranchFormat,
            errors.UnstackableRepositoryFormat,
            errors.NotStacked,
        ):
            push_result.stacked_on = None
        push_result.target_branch = br_to
        push_result.old_revid = _mod_revision.NULL_REVISION
        push_result.old_revno = 0
        # Remembers if asked explicitly or no previous location is set
        if remember or (remember is None and br_from.get_push_location() is None):
            # FIXME: Should be done only if we succeed ? -- vila 2012-01-18
            br_from.set_push_location(br_to.base)
    else:
        if stacked_on is not None:
            warning(
                "Ignoring request for a stacked branch as repository "
                "already exists at the destination location."
            )
        try:
            push_result = dir_to.push_branch(
                br_from, revision_id, overwrite, remember, create_prefix, lossy=lossy
            )
        except errors.DivergedBranches:
            raise errors.CommandError(
                gettext(
                    "These branches have diverged."
                    '  See "brz help diverged-branches"'
                    " for more information."
                )
            )
        except errors.NoRoundtrippingSupport as e:
            raise errors.CommandError(
                gettext(
                    "It is not possible to losslessly "
                    "push to %s. You may want to use --lossy."
                )
                % e.target_branch.mapping.vcs.abbreviation
            )
        except errors.NoRepositoryPresent:
            # we have a controldir but no branch or repository
            # XXX: Figure out what to do other than complain.
            raise errors.CommandError(
                gettext(
                    "At %s you have a valid .bzr"
                    " control directory, but not a branch or repository. This"
                    " is an unsupported configuration. Please move the target"
                    " directory out of the way and try again."
                )
                % location
            )
        if push_result.workingtree_updated is False:
            warning(
                "This transport does not update the working "
                "tree of: {}. See 'brz help working-trees' for "
                "more information.".format(push_result.target_branch.base)
            )
    push_result.report(to_file)
    if verbose:
        br_to = push_result.target_branch
        with br_to.lock_read():
            from .log import show_branch_change

            show_branch_change(
                br_to, to_file, push_result.old_revno, push_result.old_revid
            )



# Copyright (C) 2007-2020 Jelmer Vernoij <jelmer@jelmer.uk>
# Copyright (C) 2006, 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""The basic test suite for bzr."""

from ... import tests

TestCase = tests.TestCase
TestCaseInTempDir = tests.TestCaseInTempDir
TestCaseWithTransport = tests.TestCaseWithTransport
TestCaseWithMemoryTransport = tests.TestCaseWithMemoryTransport


def load_tests(loader, basic_tests, pattern):
    suite = loader.suiteClass()
    # add the tests for this module
    suite.addTests(basic_tests)

    prefix = __name__ + "."

    testmod_names = [
        "blackbox",
        "test_dirstate",
        "per_bzrdir",
        "per_inventory",
        "per_pack_repository",
        "per_repository_chk",
        "per_repository_vf",
        "per_versionedfile",
        "test__btree_serializer",
        "test__chk_map",
        "test__dirstate_helpers",
        "test__groupcompress",
        "test__rio",
        "test__simple_set",
        "test__static_tuple",
        "test_btree_index",
        "test_bundle",
        "test_bzrdir",
        "test_chk_map",
        "test_chk_serializer",
        "test_conflicts",
        "test_generate_ids",
        "test_groupcompress",
        "test_hashcache",
        "test_index",
        "test_inv",
        "test_inventory_delta",
        "test_knit",
        "test_matchers",
        "test_pack",
        "test_read_bundle",
        "test_remote",
        "test_repository",
        "test_rio",
        "test_smart",
        "test_smart_request",
        "test_smart_signals",
        "test_smart_transport",
        "test_serializer",
        "test_tag",
        "test_testament",
        "test_tuned_gzip",
        "test_transform",
        "test_versionedfile",
        "test_vf_search",
        "test_vfs_ratchet",
        "test_workingtree",
        "test_workingtree_4",
        "test_weave",
        "test_xml",
    ]

    # add the tests for the sub modules
    suite.addTests(
        loader.loadTestsFromModuleNames(
            [prefix + module_name for module_name in testmod_names]
        )
    )
    return suite



# Copyright (C) 2005-2009, 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from .. import missing, tests
from ..missing import iter_log_revisions
from . import TestCaseWithTransport


class TestMissing(TestCaseWithTransport):
    def assertUnmerged(self, expected, source, target, restrict="all", backward=False):
        unmerged = missing.find_unmerged(
            source, target, restrict=restrict, backward=backward
        )
        self.assertEqual(expected, unmerged)

    def test_find_unmerged(self):
        original_tree = self.make_branch_and_tree("original")
        original = original_tree.branch
        puller_tree = self.make_branch_and_tree("puller")
        puller = puller_tree.branch
        merger_tree = self.make_branch_and_tree("merger")
        merger = merger_tree.branch
        self.assertUnmerged(([], []), original, puller)
        original_tree.commit("a", rev_id=b"a")
        self.assertUnmerged(([("1", b"a", 0)], []), original, puller)
        puller_tree.pull(original)
        self.assertUnmerged(([], []), original, puller)
        merger_tree.pull(original)
        original_tree.commit("b", rev_id=b"b")
        original_tree.commit("c", rev_id=b"c")
        self.assertUnmerged(([("2", b"b", 0), ("3", b"c", 0)], []), original, puller)
        self.assertUnmerged(
            ([("3", b"c", 0), ("2", b"b", 0)], []), original, puller, backward=True
        )

        puller_tree.pull(original)
        self.assertUnmerged(([], []), original, puller)
        self.assertUnmerged(([("2", b"b", 0), ("3", b"c", 0)], []), original, merger)
        merger_tree.merge_from_branch(original)
        self.assertUnmerged(([("2", b"b", 0), ("3", b"c", 0)], []), original, merger)
        merger_tree.commit("d", rev_id=b"d")
        self.assertUnmerged(([], [("2", b"d", 0)]), original, merger)

    def test_iter_log_revisions(self):
        base_tree = self.make_branch_and_tree("base")
        self.build_tree(["base/a"])
        base_tree.add(["a"], ids=[b"a-id"])
        base_tree.commit("add a", rev_id=b"b-1")

        child_tree = base_tree.controldir.sprout("child").open_workingtree()

        self.build_tree(["child/b"])
        child_tree.add(["b"], ids=[b"b-id"])
        child_tree.commit("adding b", rev_id=b"c-2")

        child_tree.remove(["a"])
        child_tree.commit("removing a", rev_id=b"c-3")

        self.build_tree_contents([("child/b", b"new contents for b\n")])
        child_tree.commit("modifying b", rev_id=b"c-4")

        child_tree.rename_one("b", "c")
        child_tree.commit("rename b=>c", rev_id=b"c-5")

        base_extra, child_extra = missing.find_unmerged(
            base_tree.branch, child_tree.branch
        )
        results = list(
            iter_log_revisions(base_extra, base_tree.branch.repository, verbose=True)
        )
        self.assertEqual([], results)

        results = list(
            iter_log_revisions(child_extra, child_tree.branch.repository, verbose=True)
        )
        self.assertEqual(4, len(results))

        r0, r1, r2, r3 = results

        self.assertEqual(
            [
                ("2", b"c-2"),
                ("3", b"c-3"),
                ("4", b"c-4"),
                ("5", b"c-5"),
            ],
            [(r.revno, r.rev.revision_id) for r in results],
        )

        delta0 = r0.delta
        self.assertNotEqual(None, delta0)
        self.assertEqual(
            [("b", "file")], [(c.path[1], c.kind[1]) for c in delta0.added]
        )
        self.assertEqual([], delta0.removed)
        self.assertEqual([], delta0.renamed)
        self.assertEqual([], delta0.copied)
        self.assertEqual([], delta0.modified)

        delta1 = r1.delta
        self.assertNotEqual(None, delta1)
        self.assertEqual([], delta1.added)
        self.assertEqual(
            [("a", "file")], [(c.path[0], c.kind[0]) for c in delta1.removed]
        )
        self.assertEqual([], delta1.renamed)
        self.assertEqual([], delta1.copied)
        self.assertEqual([], delta1.modified)

        delta2 = r2.delta
        self.assertNotEqual(None, delta2)
        self.assertEqual([], delta2.added)
        self.assertEqual([], delta2.removed)
        self.assertEqual([], delta2.renamed)
        self.assertEqual([], delta2.copied)
        self.assertEqual(
            [("b", "file", True, False)],
            [
                (c.path[1], c.kind[1], c.changed_content, c.meta_modified())
                for c in delta2.modified
            ],
        )

        delta3 = r3.delta
        self.assertNotEqual(None, delta3)
        self.assertEqual([], delta3.added)
        self.assertEqual([], delta3.copied)
        self.assertEqual([], delta3.removed)
        self.assertEqual(
            [("b", "c", "file", False, False)],
            [
                (c.path[0], c.path[1], c.kind[1], c.changed_content, c.meta_modified())
                for c in delta3.renamed
            ],
        )
        self.assertEqual([], delta3.modified)


class TestFindUnmerged(tests.TestCaseWithTransport):
    def assertUnmerged(
        self,
        local,
        remote,
        local_branch,
        remote_branch,
        restrict="all",
        include_merged=False,
        backward=False,
        local_revid_range=None,
        remote_revid_range=None,
    ):
        """Check the output of find_unmerged_mainline_revisions."""
        local_extra, remote_extra = missing.find_unmerged(
            local_branch,
            remote_branch,
            restrict,
            include_merged=include_merged,
            backward=backward,
            local_revid_range=local_revid_range,
            remote_revid_range=remote_revid_range,
        )
        self.assertEqual(local, local_extra)
        self.assertEqual(remote, remote_extra)

    def test_same_branch(self):
        tree = self.make_branch_and_tree("tree")
        rev1 = tree.commit("one")
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertUnmerged([], [], tree.branch, tree.branch)
        self.assertUnmerged(
            [], [], tree.branch, tree.branch, local_revid_range=(rev1, rev1)
        )

    def test_one_ahead(self):
        tree = self.make_branch_and_tree("tree")
        tree.commit("one")
        tree2 = tree.controldir.sprout("tree2").open_workingtree()
        rev2 = tree2.commit("two")
        self.assertUnmerged([], [("2", rev2, 0)], tree.branch, tree2.branch)
        self.assertUnmerged([("2", rev2, 0)], [], tree2.branch, tree.branch)

    def test_restrict(self):
        tree = self.make_branch_and_tree("tree")
        tree.commit("one")
        tree2 = tree.controldir.sprout("tree2").open_workingtree()
        rev2 = tree2.commit("two")
        self.assertUnmerged([], [("2", rev2, 0)], tree.branch, tree2.branch)
        self.assertUnmerged([], None, tree.branch, tree2.branch, "local")
        self.assertUnmerged(None, [("2", rev2, 0)], tree.branch, tree2.branch, "remote")

    def test_merged(self):
        tree = self.make_branch_and_tree("tree")
        rev1 = tree.commit("one")
        tree2 = tree.controldir.sprout("tree2").open_workingtree()
        tree2.commit("two")
        tree2.commit("three")
        tree.merge_from_branch(tree2.branch)
        rev4 = tree.commit("four")

        self.assertUnmerged([("2", rev4, 0)], [], tree.branch, tree2.branch)
        self.assertUnmerged(
            [("2", rev4, 0)],
            [],
            tree.branch,
            tree2.branch,
            local_revid_range=(rev4, rev4),
        )
        self.assertUnmerged(
            [], [], tree.branch, tree2.branch, local_revid_range=(rev1, rev1)
        )

    def test_include_merged(self):
        tree = self.make_branch_and_tree("tree")
        tree.commit("one", rev_id=b"rev1")

        tree2 = tree.controldir.sprout("tree2").open_workingtree()
        tree2.commit("two", rev_id=b"rev2")
        rev3 = tree2.commit("three", rev_id=b"rev3")

        tree3 = tree2.controldir.sprout("tree3").open_workingtree()
        rev4 = tree3.commit("four", rev_id=b"rev4")
        rev5 = tree3.commit("five", rev_id=b"rev5")

        tree2.merge_from_branch(tree3.branch)
        rev6 = tree2.commit("six", rev_id=b"rev6")

        self.assertUnmerged(
            [],
            [
                ("2", b"rev2", 0),
                ("3", b"rev3", 0),
                ("4", b"rev6", 0),
                ("3.1.1", b"rev4", 1),
                ("3.1.2", b"rev5", 1),
            ],
            tree.branch,
            tree2.branch,
            include_merged=True,
        )

        self.assertUnmerged(
            [],
            [
                ("4", b"rev6", 0),
                ("3.1.2", b"rev5", 1),
                ("3.1.1", b"rev4", 1),
                ("3", b"rev3", 0),
                ("2", b"rev2", 0),
            ],
            tree.branch,
            tree2.branch,
            include_merged=True,
            backward=True,
        )

        self.assertUnmerged(
            [],
            [("4", b"rev6", 0)],
            tree.branch,
            tree2.branch,
            include_merged=True,
            remote_revid_range=(rev6, rev6),
        )

        self.assertUnmerged(
            [],
            [("3", b"rev3", 0), ("3.1.1", b"rev4", 1)],
            tree.branch,
            tree2.branch,
            include_merged=True,
            remote_revid_range=(rev3, rev4),
        )

        self.assertUnmerged(
            [],
            [("4", b"rev6", 0), ("3.1.2", b"rev5", 1)],
            tree.branch,
            tree2.branch,
            include_merged=True,
            remote_revid_range=(rev5, rev6),
        )

    def test_revision_range(self):
        local = self.make_branch_and_tree("local")
        lrevid1 = local.commit("one")
        remote = local.controldir.sprout("remote").open_workingtree()
        rrevid2 = remote.commit("two")
        rrevid3 = remote.commit("three")
        rrevid4 = remote.commit("four")
        lrevid2 = local.commit("two")
        lrevid3 = local.commit("three")
        lrevid4 = local.commit("four")
        local_extra = [("2", lrevid2, 0), ("3", lrevid3, 0), ("4", lrevid4, 0)]
        remote_extra = [("2", rrevid2, 0), ("3", rrevid3, 0), ("4", rrevid4, 0)]

        # control
        self.assertUnmerged(local_extra, remote_extra, local.branch, remote.branch)
        self.assertUnmerged(
            local_extra,
            remote_extra,
            local.branch,
            remote.branch,
            local_revid_range=(None, None),
            remote_revid_range=(None, None),
        )

        # exclude local revisions
        self.assertUnmerged(
            [("2", lrevid2, 0)],
            remote_extra,
            local.branch,
            remote.branch,
            local_revid_range=(lrevid2, lrevid2),
        )
        self.assertUnmerged(
            [("2", lrevid2, 0), ("3", lrevid3, 0)],
            remote_extra,
            local.branch,
            remote.branch,
            local_revid_range=(lrevid2, lrevid3),
        )
        self.assertUnmerged(
            [("2", lrevid2, 0), ("3", lrevid3, 0)],
            None,
            local.branch,
            remote.branch,
            "local",
            local_revid_range=(lrevid2, lrevid3),
        )

        # exclude remote revisions
        self.assertUnmerged(
            local_extra,
            [("2", rrevid2, 0)],
            local.branch,
            remote.branch,
            remote_revid_range=(None, rrevid2),
        )
        self.assertUnmerged(
            local_extra,
            [("2", rrevid2, 0)],
            local.branch,
            remote.branch,
            remote_revid_range=(lrevid1, rrevid2),
        )
        self.assertUnmerged(
            local_extra,
            [("2", rrevid2, 0)],
            local.branch,
            remote.branch,
            remote_revid_range=(rrevid2, rrevid2),
        )
        self.assertUnmerged(
            local_extra,
            [("2", rrevid2, 0), ("3", rrevid3, 0)],
            local.branch,
            remote.branch,
            remote_revid_range=(None, rrevid3),
        )
        self.assertUnmerged(
            local_extra,
            [("2", rrevid2, 0), ("3", rrevid3, 0)],
            local.branch,
            remote.branch,
            remote_revid_range=(rrevid2, rrevid3),
        )
        self.assertUnmerged(
            local_extra,
            [("3", rrevid3, 0)],
            local.branch,
            remote.branch,
            remote_revid_range=(rrevid3, rrevid3),
        )
        self.assertUnmerged(
            None,
            [("2", rrevid2, 0), ("3", rrevid3, 0)],
            local.branch,
            remote.branch,
            "remote",
            remote_revid_range=(rrevid2, rrevid3),
        )

        # exclude local and remote revisions
        self.assertUnmerged(
            [("3", lrevid3, 0)],
            [("3", rrevid3, 0)],
            local.branch,
            remote.branch,
            local_revid_range=(lrevid3, lrevid3),
            remote_revid_range=(rrevid3, rrevid3),
        )



# Copyright (C) 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test that all trees support Tree.list_files()."""

from breezy import osutils
from breezy.tests import TestNotApplicable
from breezy.tests.per_tree import TestCaseWithTree


class TestListFiles(TestCaseWithTree):
    def assertFilesListEqual(self, tree, expected, **kwargs):
        with tree.lock_read():
            if tree.supports_file_ids:
                actual = [
                    (path, status, kind, ie.file_id)
                    for path, status, kind, ie in tree.list_files(**kwargs)
                ]
                expected = [
                    (
                        path,
                        status,
                        kind,
                        tree.path2id(
                            osutils.pathjoin(kwargs.get("from_dir", ""), path)
                        ),
                    )
                    for path, status, kind in expected
                ]
            else:
                actual = [
                    (path, status, kind)
                    for path, status, kind, ie in tree.list_files(**kwargs)
                ]
                expected = [(path, status, kind) for path, status, kind in expected]
        self.assertEqual(expected, actual)

    def test_list_files_with_root(self):
        work_tree = self.make_branch_and_tree("wt")
        tree = self.get_tree_no_parents_abc_content(work_tree)
        expected = [
            ("", "V", "directory"),
            ("a", "V", "file"),
            ("b", "V", "directory"),
            ("b/c", "V", "file"),
        ]

        self.assertFilesListEqual(tree, expected, include_root=True)

    def test_list_files_no_root(self):
        work_tree = self.make_branch_and_tree("wt")
        tree = self.get_tree_no_parents_abc_content(work_tree)
        expected = [
            ("a", "V", "file"),
            ("b", "V", "directory"),
            ("b/c", "V", "file"),
        ]
        self.assertFilesListEqual(tree, expected)

    def test_list_files_with_root_no_recurse(self):
        work_tree = self.make_branch_and_tree("wt")
        tree = self.get_tree_no_parents_abc_content(work_tree)
        expected = [
            ("", "V", "directory"),
            ("a", "V", "file"),
        ]
        expected.append(("b", "V", "directory"))
        self.assertFilesListEqual(tree, expected, include_root=True, recursive=False)

    def test_list_files_no_root_no_recurse(self):
        work_tree = self.make_branch_and_tree("wt")
        tree = self.get_tree_no_parents_abc_content(work_tree)
        expected = [("a", "V", "file")]
        expected.append(("b", "V", "directory"))
        self.assertFilesListEqual(tree, expected, recursive=False)

    def test_list_files_from_dir(self):
        work_tree = self.make_branch_and_tree("wt")
        tree = self.get_tree_no_parents_abc_content(work_tree)
        expected = [("c", "V", "file")]
        self.assertFilesListEqual(tree, expected, from_dir="b")

    def test_list_files_from_dir_no_recurse(self):
        # The test trees don't have much nesting so test with an explicit root
        work_tree = self.make_branch_and_tree("wt")
        tree = self.get_tree_no_parents_abc_content(work_tree)
        expected = [("a", "V", "file")]
        expected.append(("b", "V", "directory"))

        self.assertFilesListEqual(tree, expected, from_dir="", recursive=False)

    def skip_if_no_reference(self, tree):
        if not getattr(tree, "supports_tree_reference", lambda: False)():
            raise TestNotApplicable("Tree references not supported")

    def create_nested(self):
        work_tree = self.make_branch_and_tree("wt")
        with work_tree.lock_write():
            self.skip_if_no_reference(work_tree)
            subtree = self.make_branch_and_tree("wt/subtree")
            self.build_tree(["wt/subtree/a"])
            subtree.add(["a"])
            subtree.commit("foo")
            work_tree.add_reference(subtree)
        tree = self._convert_tree(work_tree)
        self.skip_if_no_reference(tree)
        return tree, subtree

    def test_list_files_with_unfollowed_reference(self):
        tree, subtree = self.create_nested()
        expected = [("", "V", "directory"), ("subtree", "V", "tree-reference")]
        self.assertFilesListEqual(
            tree, expected, recursive=True, recurse_nested=False, include_root=True
        )

    def test_list_files_with_followed_reference(self):
        tree, subtree = self.create_nested()
        expected = [
            ("", "V", "directory"),
            ("subtree", "V", "directory"),
            ("subtree/a", "V", "file"),
        ]
        self.assertFilesListEqual(
            tree, expected, recursive=True, recurse_nested=True, include_root=True
        )



# Copyright (C) 2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import sys
import time

from breezy import errors, tests
from breezy import revision as _mod_revision
from breezy.bzr import inventory, remote
from breezy.bzr.tests.per_repository_vf import (
    TestCaseWithRepository,
    all_repository_vf_format_scenarios,
)
from breezy.tests.scenarios import load_tests_apply_scenarios

load_tests = load_tests_apply_scenarios


class FileIdInvolvedWGhosts(TestCaseWithRepository):
    scenarios = all_repository_vf_format_scenarios()

    def create_branch_with_ghost_text(self):
        builder = self.make_branch_builder("ghost")
        builder.build_snapshot(
            None,
            [
                ("add", ("", b"root-id", "directory", None)),
                ("add", ("a", b"a-file-id", "file", b"some content\n")),
            ],
            revision_id=b"A-id",
        )
        b = builder.get_branch()
        old_rt = b.repository.revision_tree(b"A-id")
        new_inv = inventory.mutable_inventory_from_tree(old_rt)
        new_inv.revision_id = b"B-id"
        new_inv.get_entry(b"a-file-id").revision = b"ghost-id"
        new_rev = _mod_revision.Revision(
            b"B-id",
            timestamp=time.time(),
            timezone=0,
            message="Committing against a ghost",
            committer="Joe Foo <joe@foo.com>",
            properties={},
            parent_ids=(b"A-id", b"ghost-id"),
        )
        b.lock_write()
        self.addCleanup(b.unlock)
        b.repository.start_write_group()
        b.repository.add_revision(b"B-id", new_rev, new_inv)
        self.disable_commit_write_group_paranoia(b.repository)
        b.repository.commit_write_group()
        return b

    def disable_commit_write_group_paranoia(self, repo):
        if isinstance(repo, remote.RemoteRepository):
            # We can't easily disable the checks in a remote repo.
            repo.abort_write_group()
            raise tests.TestSkipped(
                "repository format does not support storing revisions with "
                "missing texts."
            )
        pack_coll = getattr(repo, "_pack_collection", None)
        if pack_coll is not None:
            # Monkey-patch the pack collection instance to allow storing
            # incomplete revisions.
            pack_coll._check_new_inventories = lambda: []

    def test_file_ids_include_ghosts(self):
        b = self.create_branch_with_ghost_text()
        repo = b.repository
        self.assertEqual(
            {b"a-file-id": {b"ghost-id"}},
            repo.fileids_altered_by_revision_ids([b"B-id"]),
        )

    def test_file_ids_uses_fallbacks(self):
        builder = self.make_branch_builder("source", format=self.bzrdir_format)
        repo = builder.get_branch().repository
        if not repo._format.supports_external_lookups:
            raise tests.TestNotApplicable("format does not support stacking")
        builder.start_series()
        builder.build_snapshot(
            None,
            [
                ("add", ("", b"root-id", "directory", None)),
                ("add", ("file", b"file-id", "file", b"contents\n")),
            ],
            revision_id=b"A-id",
        )
        builder.build_snapshot(
            [b"A-id"], [("modify", ("file", b"new-content\n"))], revision_id=b"B-id"
        )
        builder.build_snapshot(
            [b"B-id"],
            [("modify", ("file", b"yet more content\n"))],
            revision_id=b"C-id",
        )
        builder.finish_series()
        source_b = builder.get_branch()
        source_b.lock_read()
        self.addCleanup(source_b.unlock)
        base = self.make_branch("base")
        base.pull(source_b, stop_revision=b"B-id")
        stacked = self.make_branch("stacked")
        stacked.set_stacked_on_url("../base")
        stacked.pull(source_b, stop_revision=b"C-id")

        stacked.lock_read()
        self.addCleanup(stacked.unlock)
        repo = stacked.repository
        keys = {b"file-id": {b"A-id"}}
        if stacked.repository.supports_rich_root():
            keys[b"root-id"] = {b"A-id"}
        self.assertEqual(keys, repo.fileids_altered_by_revision_ids([b"A-id"]))


class FileIdInvolvedBase(TestCaseWithRepository):
    def touch(self, tree, filename):
        # use the trees transport to not depend on the tree's location or type.
        tree.controldir.root_transport.append_bytes(filename, b"appended line\n")

    def compare_tree_fileids(self, branch, old_rev, new_rev):
        old_tree = self.branch.repository.revision_tree(old_rev)
        new_tree = self.branch.repository.revision_tree(new_rev)
        delta = new_tree.changes_from(old_tree)

        l2 = (
            [change.file_id for change in delta.added]
            + [change.file_id for change in delta.renamed]
            + [change.file_id for change in delta.modified]
            + [change.file_id for change in delta.copied]
        )
        return set(l2)


class TestFileIdInvolved(FileIdInvolvedBase):
    scenarios = all_repository_vf_format_scenarios()

    def setUp(self):
        super().setUp()
        # create three branches, and merge it
        #
        #          ,-->J------>K                (branch2)
        #         /             \
        #  A --->B --->C---->D-->G              (main)
        #  \          /     /
        #   '--->E---+---->F                    (branch1)

        # A changes:
        # B changes: 'a-file-id-2006-01-01-abcd'
        # C changes:  Nothing (perfect merge)
        # D changes: 'b-file-id-2006-01-01-defg'
        # E changes: 'file-d'
        # F changes: 'file-d'
        # G changes: 'b-file-id-2006-01-01-defg'
        # J changes: 'b-file-id-2006-01-01-defg'
        # K changes: 'c-funky<file-id>quiji%bo'

        main_wt = self.make_branch_and_tree("main")
        main_branch = main_wt.branch
        self.build_tree(["main/a", "main/b", "main/c"])

        main_wt.add(
            ["a", "b", "c"],
            ids=[
                b"a-file-id-2006-01-01-abcd",
                b"b-file-id-2006-01-01-defg",
                b"c-funky<file-id>quiji%bo",
            ],
        )
        try:
            main_wt.commit("Commit one", rev_id=b"rev-A")
        except errors.IllegalPath as e:
            # TODO: jam 20060701 Consider raising a different exception
            #       newer formats do support this, and nothin can done to
            #       correct this test - its not a bug.
            if sys.platform == "win32":
                raise tests.TestSkipped(
                    "Old repository formats do not support file ids with <> on win32"
                ) from e
            # This is not a known error condition
            raise

        # -------- end A -----------

        bt1 = self.make_branch_and_tree("branch1")
        bt1.pull(main_branch)
        b1 = bt1.branch
        self.build_tree(["branch1/d"])
        bt1.add(["d"], ids=[b"file-d"])
        bt1.commit("branch1, Commit one", rev_id=b"rev-E")

        # -------- end E -----------

        self.touch(main_wt, "a")
        main_wt.commit("Commit two", rev_id=b"rev-B")

        # -------- end B -----------

        bt2 = self.make_branch_and_tree("branch2")
        bt2.pull(main_branch)
        branch2_branch = bt2.branch
        set_executability(bt2, "b", True)
        bt2.commit("branch2, Commit one", rev_id=b"rev-J")

        # -------- end J -----------

        main_wt.merge_from_branch(b1)
        main_wt.commit("merge branch1, rev-11", rev_id=b"rev-C")

        # -------- end C -----------

        bt1.rename_one("d", "e")
        bt1.commit("branch1, commit two", rev_id=b"rev-F")

        # -------- end F -----------

        self.touch(bt2, "c")
        bt2.commit("branch2, commit two", rev_id=b"rev-K")

        # -------- end K -----------

        main_wt.merge_from_branch(b1)
        self.touch(main_wt, "b")
        # D gets some funky characters to make sure the unescaping works
        main_wt.commit("merge branch1, rev-12", rev_id=b"rev-<D>")

        # end D

        main_wt.merge_from_branch(branch2_branch)
        main_wt.commit("merge branch1, rev-22", rev_id=b"rev-G")

        # end G
        self.branch = main_branch

    def test_fileids_altered_between_two_revs(self):
        self.branch.lock_read()
        self.addCleanup(self.branch.unlock)
        self.branch.repository.fileids_altered_by_revision_ids([b"rev-J", b"rev-K"])
        self.assertEqual(
            {
                b"b-file-id-2006-01-01-defg": {b"rev-J"},
                b"c-funky<file-id>quiji%bo": {b"rev-K"},
            },
            self.branch.repository.fileids_altered_by_revision_ids(
                [b"rev-J", b"rev-K"]
            ),
        )

        self.assertEqual(
            {
                b"b-file-id-2006-01-01-defg": {b"rev-<D>"},
                b"file-d": {b"rev-F"},
            },
            self.branch.repository.fileids_altered_by_revision_ids(
                [b"rev-<D>", b"rev-F"]
            ),
        )

        self.assertEqual(
            {
                b"b-file-id-2006-01-01-defg": {b"rev-<D>", b"rev-G", b"rev-J"},
                b"c-funky<file-id>quiji%bo": {b"rev-K"},
                b"file-d": {b"rev-F"},
            },
            self.branch.repository.fileids_altered_by_revision_ids(
                [b"rev-<D>", b"rev-G", b"rev-F", b"rev-K", b"rev-J"]
            ),
        )

        self.assertEqual(
            {
                b"a-file-id-2006-01-01-abcd": {b"rev-B"},
                b"b-file-id-2006-01-01-defg": {b"rev-<D>", b"rev-G", b"rev-J"},
                b"c-funky<file-id>quiji%bo": {b"rev-K"},
                b"file-d": {b"rev-F"},
            },
            self.branch.repository.fileids_altered_by_revision_ids(
                [b"rev-G", b"rev-F", b"rev-C", b"rev-B", b"rev-<D>", b"rev-K", b"rev-J"]
            ),
        )

    def fileids_altered_by_revision_ids(self, revision_ids):
        """This is a wrapper to strip TREE_ROOT if it occurs."""
        repo = self.branch.repository
        root_id = self.branch.basis_tree().path2id("")
        result = repo.fileids_altered_by_revision_ids(revision_ids)
        if root_id in result:
            del result[root_id]
        return result

    def test_fileids_altered_by_revision_ids(self):
        self.branch.lock_read()
        self.addCleanup(self.branch.unlock)
        self.assertEqual(
            {
                b"a-file-id-2006-01-01-abcd": {b"rev-A"},
                b"b-file-id-2006-01-01-defg": {b"rev-A"},
                b"c-funky<file-id>quiji%bo": {b"rev-A"},
            },
            self.fileids_altered_by_revision_ids([b"rev-A"]),
        )
        self.assertEqual(
            {b"a-file-id-2006-01-01-abcd": {b"rev-B"}},
            self.branch.repository.fileids_altered_by_revision_ids([b"rev-B"]),
        )
        self.assertEqual(
            {b"b-file-id-2006-01-01-defg": {b"rev-<D>"}},
            self.branch.repository.fileids_altered_by_revision_ids([b"rev-<D>"]),
        )

    def test_fileids_involved_full_compare(self):
        # this tests that the result of each fileid_involved calculation
        # along a revision history selects only the fileids selected by
        # comparing the trees - no less, and no more. This is correct
        # because in our sample data we do not revert any file ids along
        # the revision history.
        self.branch.lock_read()
        self.addCleanup(self.branch.unlock)
        graph = self.branch.repository.get_graph()
        history = list(
            graph.iter_lefthand_ancestry(
                self.branch.last_revision(), [_mod_revision.NULL_REVISION]
            )
        )
        history.reverse()

        if len(history) < 2:
            return

        for start in range(0, len(history) - 1):
            start_id = history[start]
            for end in range(start + 1, len(history)):
                end_id = history[end]
                unique_revs = graph.find_unique_ancestors(end_id, [start_id])
                l1 = self.branch.repository.fileids_altered_by_revision_ids(unique_revs)
                l1 = set(l1.keys())
                l2 = self.compare_tree_fileids(self.branch, start_id, end_id)
                self.assertEqual(l1, l2)


class TestFileIdInvolvedNonAscii(FileIdInvolvedBase):
    scenarios = all_repository_vf_format_scenarios()

    def test_utf8_file_ids_and_revision_ids(self):
        main_wt = self.make_branch_and_tree("main")
        self.build_tree(["main/a"])

        file_id = "a-f\xedle-id".encode()
        main_wt.add(["a"], ids=[file_id])
        revision_id = "r\xe9v-a".encode()
        try:
            main_wt.commit("a", rev_id=revision_id)
        except errors.NonAsciiRevisionId as e:
            raise tests.TestSkipped(
                "non-ascii revision ids not supported by {}".format(
                    self.repository_format
                )
            ) from e

        repo = main_wt.branch.repository
        repo.lock_read()
        self.addCleanup(repo.unlock)
        file_ids = repo.fileids_altered_by_revision_ids([revision_id])
        root_id = main_wt.basis_tree().path2id("")
        if root_id in file_ids:
            self.assertEqual({file_id: {revision_id}, root_id: {revision_id}}, file_ids)
        else:
            self.assertEqual({file_id: {revision_id}}, file_ids)


class TestFileIdInvolvedSuperset(FileIdInvolvedBase):
    scenarios = all_repository_vf_format_scenarios()

    def setUp(self):
        super().setUp()

        self.branch = None
        main_wt = self.make_branch_and_tree("main")
        main_branch = main_wt.branch
        self.build_tree(["main/a", "main/b", "main/c"])

        main_wt.add(
            ["a", "b", "c"],
            ids=[
                b"a-file-id-2006-01-01-abcd",
                b"b-file-id-2006-01-01-defg",
                b"c-funky<file-id>quiji'\"%bo",
            ],
        )
        try:
            main_wt.commit("Commit one", rev_id=b"rev-A")
        except errors.IllegalPath as e:
            # TODO: jam 20060701 Consider raising a different exception
            #       newer formats do support this, and nothin can done to
            #       correct this test - its not a bug.
            if sys.platform == "win32":
                raise tests.TestSkipped(
                    "Old repository formats do not support file ids with <> on win32"
                ) from e
            # This is not a known error condition
            raise

        branch2_wt = self.make_branch_and_tree("branch2")
        branch2_wt.pull(main_branch)
        branch2_bzrdir = branch2_wt.controldir
        branch2_branch = branch2_bzrdir.open_branch()
        set_executability(branch2_wt, "b", True)
        branch2_wt.commit("branch2, Commit one", rev_id=b"rev-J")

        main_wt.merge_from_branch(branch2_branch)
        set_executability(main_wt, "b", False)
        main_wt.commit("merge branch1, rev-22", rev_id=b"rev-G")

        # end G
        self.branch = main_branch

    def test_fileid_involved_full_compare2(self):
        # this tests that fileids_altered_by_revision_ids returns
        # more information than compare_tree can, because it
        # sees each change rather than the aggregate delta.
        self.branch.lock_read()
        self.addCleanup(self.branch.unlock)
        graph = self.branch.repository.get_graph()
        history = list(
            graph.iter_lefthand_ancestry(
                self.branch.last_revision(), [_mod_revision.NULL_REVISION]
            )
        )
        history.reverse()
        old_rev = history[0]
        new_rev = history[1]
        unique_revs = graph.find_unique_ancestors(new_rev, [old_rev])

        l1 = self.branch.repository.fileids_altered_by_revision_ids(unique_revs)
        l1 = set(l1.keys())

        l2 = self.compare_tree_fileids(self.branch, old_rev, new_rev)
        self.assertNotEqual(l2, l1)
        self.assertSubset(l2, l1)


def set_executability(wt, path, executable=True):
    """Set the executable bit for the file at path in the working tree.

    os.chmod() doesn't work on windows. But TreeTransform can mark or
    unmark a file as executable.
    """
    with wt.transform() as tt:
        tt.set_executability(executable, tt.trans_id_tree_path(path))
        tt.apply()



# Copyright (C) 2006-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Tests of the brz serve command."""

import signal
import sys
import threading
from _thread import interrupt_main  # type: ignore

from ... import builtins, config, errors, osutils, trace, transport, urlutils
from ... import revision as _mod_revision
from ...branch import Branch
from ...bzr.smart import client, medium
from ...bzr.smart.server import BzrServerFactory, SmartTCPServer
from ...controldir import ControlDir
from ...transport import remote
from .. import TestCaseWithMemoryTransport, TestCaseWithTransport


class TestBzrServeBase(TestCaseWithTransport):
    def run_bzr_serve_then_func(
        self, serve_args, retcode=0, func=None, *func_args, **func_kwargs
    ):
        """Run 'brz serve', and run the given func in a thread once the server
        has started.

        When 'func' terminates, the server will be terminated too.

        Returns stdout and stderr.
        """

        def on_server_start_thread(tcp_server):
            """This runs concurrently with the server thread.

            The server is interrupted as soon as ``func`` finishes, even if an
            exception is encountered.
            """
            try:
                # Run func if set
                self.tcp_server = tcp_server
                if func is not None:
                    try:
                        func(*func_args, **func_kwargs)
                    except Exception as e:
                        # Log errors to make some test failures a little less
                        # mysterious.
                        trace.mutter("func broke: %r", e)
            finally:
                # Then stop the server
                trace.mutter("interrupting...")
                interrupt_main()

        # When the hook is fired, it just starts ``on_server_start_thread`` and
        # return

        def on_server_start(backing_urls, tcp_server):
            t = threading.Thread(target=on_server_start_thread, args=(tcp_server,))
            t.start()

        # install hook
        SmartTCPServer.hooks.install_named_hook(
            "server_started_ex", on_server_start, "run_bzr_serve_then_func hook"
        )
        # It seems interrupt_main() will not raise KeyboardInterrupt
        # until after socket.accept returns. So we set the timeout low to make
        # the test faster.
        self.overrideAttr(SmartTCPServer, "_ACCEPT_TIMEOUT", 0.1)
        # start a TCP server
        try:
            out, err = self.run_bzr(["serve"] + list(serve_args), retcode=retcode)
        except KeyboardInterrupt:
            return (self._last_cmd_stdout.getvalue(), self._last_cmd_stderr.getvalue())
        return out, err


class TestBzrServe(TestBzrServeBase):
    def setUp(self):
        super().setUp()
        self.disable_missing_extensions_warning()

    def test_server_exception_with_hook(self):
        """Catch exception from the server in the server_exception hook.

        We use ``run_bzr_serve_then_func`` without a ``func`` so the server
        will receive a KeyboardInterrupt exception we want to catch.
        """

        def hook(exception):
            if exception[0] is KeyboardInterrupt:
                sys.stderr.write(b"catching KeyboardInterrupt\n")
                return True
            else:
                return False

        SmartTCPServer.hooks.install_named_hook(
            "server_exception", hook, "test_server_except_hook hook"
        )
        args = ["--listen", "localhost", "--port", "0", "--quiet"]
        out, err = self.run_bzr_serve_then_func(args, retcode=0)
        self.assertEqual("catching KeyboardInterrupt\n", err)

    def test_server_exception_no_hook(self):
        """Test exception without hook returns error."""
        args = []
        out, err = self.run_bzr_serve_then_func(args, retcode=3)

    def assertInetServerShutsdownCleanly(self, process):
        """Shutdown the server process looking for errors."""
        # Shutdown the server: the server should shut down when it cannot read
        # from stdin anymore.
        process.stdin.close()
        # Hide stdin from the subprocess module, so it won't fail to close it.
        process.stdin = None
        result = self.finish_brz_subprocess(process)
        self.assertEqual(b"", result[0])
        self.assertEqual(b"", result[1])

    def assertServerFinishesCleanly(self, process):
        """Shutdown the brz serve instance process looking for errors."""
        # Shutdown the server
        result = self.finish_brz_subprocess(
            process, retcode=3, send_signal=signal.SIGINT
        )
        self.assertEqual(b"", result[0])
        self.assertEqual(b"brz: interrupted\n", result[1])

    def make_read_requests(self, branch):
        """Do some read only requests."""
        with branch.lock_read():
            branch.repository.all_revision_ids()
            self.assertEqual(_mod_revision.NULL_REVISION, branch.last_revision())

    def start_server_inet(self, extra_options=()):
        """Start a brz server subprocess using the --inet option.

        :param extra_options: extra options to give the server.
        :return: a tuple with the brz process handle for passing to
            finish_brz_subprocess, a client for the server, and a transport.
        """
        # Serve from the current directory
        args = ["serve", "--inet"]
        args.extend(extra_options)
        process = self.start_brz_subprocess(args)

        # Connect to the server
        # We use this url because while this is no valid URL to connect to this
        # server instance, the transport needs a URL.
        url = "bzr://localhost/"
        self.permit_url(url)
        client_medium = medium.SmartSimplePipesClientMedium(
            process.stdout, process.stdin, url
        )
        transport = remote.RemoteTransport(url, medium=client_medium)
        return process, transport

    def start_server_port(self, extra_options=()):
        """Start a brz server subprocess.

        :param extra_options: extra options to give the server.
        :return: a tuple with the brz process handle for passing to
            finish_brz_subprocess, and the base url for the server.
        """
        # Serve from the current directory
        args = ["serve", "--listen", "localhost", "--port", "0"]
        args.extend(extra_options)
        process = self.start_brz_subprocess(args, skip_if_plan_to_signal=True)
        port_line = process.stderr.readline()
        prefix = b"listening on port: "
        self.assertStartsWith(port_line, prefix)
        port = int(port_line[len(prefix) :])
        url = "bzr://localhost:%d/" % port
        self.permit_url(url)
        return process, url

    def test_bzr_serve_quiet(self):
        self.make_branch(".")
        args = ["--listen", "localhost", "--port", "0", "--quiet"]
        out, err = self.run_bzr_serve_then_func(args, retcode=3)
        self.assertEqual("", out)
        self.assertEqual("", err)

    def test_bzr_serve_inet_readonly(self):
        """Brz server should provide a read only filesystem by default."""
        process, transport = self.start_server_inet()
        self.assertRaises(errors.TransportNotPossible, transport.mkdir, "adir")
        self.assertInetServerShutsdownCleanly(process)

    def test_bzr_serve_inet_readwrite(self):
        # Make a branch
        self.make_branch(".")

        process, transport = self.start_server_inet(["--allow-writes"])

        # We get a working branch, and can create a directory
        branch = ControlDir.open_from_transport(transport).open_branch()
        self.make_read_requests(branch)
        transport.mkdir("adir")
        self.assertInetServerShutsdownCleanly(process)

    def test_bzr_serve_port_readonly(self):
        """Brz server should provide a read only filesystem by default."""
        process, url = self.start_server_port()
        t = transport.get_transport_from_url(url)
        self.assertRaises(errors.TransportNotPossible, t.mkdir, "adir")
        self.assertServerFinishesCleanly(process)

    def test_bzr_serve_port_readwrite(self):
        # Make a branch
        self.make_branch(".")

        process, url = self.start_server_port(["--allow-writes"])

        # Connect to the server
        branch = Branch.open(url)
        self.make_read_requests(branch)
        self.assertServerFinishesCleanly(process)

    def test_bzr_serve_supports_protocol(self):
        # Make a branch
        self.make_branch(".")

        process, url = self.start_server_port(["--allow-writes", "--protocol=bzr"])

        # Connect to the server
        branch = Branch.open(url)
        self.make_read_requests(branch)
        self.assertServerFinishesCleanly(process)

    def test_bzr_serve_dhpss(self):
        # This is a smoke test that the server doesn't crash when run with
        # -Dhpss, and does drop some hpss logging to the file.
        self.make_branch(".")
        log_fname = self.test_dir + "/server.log"
        self.overrideEnv("BRZ_LOG", log_fname)
        process, transport = self.start_server_inet(["-Dhpss"])
        branch = ControlDir.open_from_transport(transport).open_branch()
        self.make_read_requests(branch)
        self.assertInetServerShutsdownCleanly(process)
        f = open(log_fname, "rb")
        content = f.read()
        f.close()
        self.assertContainsRe(content, rb"hpss request: \[[0-9-]+\]")

    def test_bzr_serve_supports_configurable_timeout(self):
        gs = config.GlobalStack()
        gs.set("serve.client_timeout", 0.2)
        # Save the config as the subprocess will use it
        gs.store.save()
        process, url = self.start_server_port()
        self.build_tree_contents([("a_file", b"contents\n")])
        # We can connect and issue a request
        t = transport.get_transport_from_url(url)
        self.assertEqual(b"contents\n", t.get_bytes("a_file"))
        # However, if we just wait for more content from the server, it will
        # eventually disconnect us.
        m = t.get_smart_medium()
        m.read_bytes(1)
        # Now, we wait for timeout to trigger
        err = process.stderr.readline()
        self.assertEqual(
            b"Connection Timeout: disconnecting client after 0.2 seconds\n", err
        )
        self.assertServerFinishesCleanly(process)

    def test_bzr_serve_supports_client_timeout(self):
        process, url = self.start_server_port(["--client-timeout=0.1"])
        self.build_tree_contents([("a_file", b"contents\n")])
        # We can connect and issue a request
        t = transport.get_transport_from_url(url)
        self.assertEqual(b"contents\n", t.get_bytes("a_file"))
        # However, if we just wait for more content from the server, it will
        # eventually disconnect us.
        # TODO: Use something like signal.alarm() so that if the server doesn't
        #       properly handle the timeout, we end up failing the test instead
        #       of hanging forever.
        m = t.get_smart_medium()
        m.read_bytes(1)
        # Now, we wait for timeout to trigger
        err = process.stderr.readline()
        self.assertEqual(
            b"Connection Timeout: disconnecting client after 0.1 seconds\n", err
        )
        self.assertServerFinishesCleanly(process)

    def test_bzr_serve_graceful_shutdown(self):
        big_contents = b"a" * 64 * 1024
        self.build_tree_contents([("bigfile", big_contents)])
        process, url = self.start_server_port(["--client-timeout=1.0"])
        t = transport.get_transport_from_url(url)
        m = t.get_smart_medium()
        c = client._SmartClient(m)
        # Start, but don't finish a response
        resp, response_handler = c.call_expecting_body(b"get", b"bigfile")
        self.assertEqual((b"ok",), resp)
        # Note: process.send_signal is a Python 2.6ism
        process.send_signal(signal.SIGHUP)
        # Wait for the server to notice the signal, and then read the actual
        # body of the response. That way we know that it is waiting for the
        # request to finish
        self.assertEqual(b"Requested to stop gracefully\n", process.stderr.readline())
        self.assertIn(
            process.stderr.readline(), (b"", b"Waiting for 1 client(s) to finish\n")
        )
        body = response_handler.read_body_bytes()
        if body != big_contents:
            self.fail('Failed to properly read the contents of "bigfile"')
        # Now that our request is finished, the medium should notice it has
        # been disconnected.
        self.assertEqual(b"", m.read_bytes(1))
        # And the server should be stopping
        self.assertEqual(0, process.wait())


class TestCmdServeChrooting(TestBzrServeBase):
    def test_serve_tcp(self):
        """'brz serve' wraps the given --directory in a ChrootServer.

        So requests that search up through the parent directories (like
        find_repositoryV3) will give "not found" responses, rather than
        InvalidURLJoin or jail break errors.
        """
        t = self.get_transport()
        t.mkdir("server-root")
        self.run_bzr_serve_then_func(
            [
                "--listen",
                "127.0.0.1",
                "--port",
                "0",
                "--directory",
                t.local_abspath("server-root"),
                "--allow-writes",
            ],
            func=self.when_server_started,
        )
        # The when_server_started method issued a find_repositoryV3 that should
        # fail with 'norepository' because there are no repositories inside the
        # --directory.
        self.assertEqual((b"norepository",), self.client_resp)

    def when_server_started(self):
        # Connect to the TCP server and issue some requests and see what comes
        # back.
        client_medium = medium.SmartTCPClientMedium(
            "127.0.0.1",
            self.tcp_server.port,
            "bzr://localhost:%d/" % (self.tcp_server.port,),
        )
        smart_client = client._SmartClient(client_medium)
        resp = smart_client.call("mkdir", "foo", "")
        resp = smart_client.call("BzrDirFormat.initialize", "foo/")
        try:
            resp = smart_client.call("BzrDir.find_repositoryV3", "foo/")
        except errors.ErrorFromSmartServer as e:
            resp = e.error_tuple
        self.client_resp = resp
        client_medium.disconnect()


class TestUserdirExpansion(TestCaseWithMemoryTransport):
    @staticmethod
    def fake_expanduser(path):
        """A simple, environment-independent, function for the duration of this
        test.

        Paths starting with a path segment of '~user' will expand to start with
        '/home/user/'.  Every other path will be unchanged.
        """
        if path.split("/", 1)[0] == "~user":
            return "/home/user" + path[len("~user") :]
        return path

    def make_test_server(self, base_path="/"):
        """Make and start a BzrServerFactory, backed by a memory transport, and
        creat '/home/user' in that transport.
        """
        bzr_server = BzrServerFactory(self.fake_expanduser, lambda t: base_path)
        mem_transport = self.get_transport()
        mem_transport.mkdir("home")
        mem_transport.mkdir("home/user")
        bzr_server.set_up(mem_transport, None, None, inet=True, timeout=4.0)
        self.addCleanup(bzr_server.tear_down)
        return bzr_server

    def test_bzr_serve_expands_userdir(self):
        bzr_server = self.make_test_server()
        self.assertTrue(bzr_server.smart_server.backing_transport.has("~user"))

    def test_bzr_serve_does_not_expand_userdir_outside_base(self):
        bzr_server = self.make_test_server("/foo")
        self.assertFalse(bzr_server.smart_server.backing_transport.has("~user"))

    def test_get_base_path(self):
        """cmd_serve will turn the --directory option into a LocalTransport
        (optionally decorated with 'readonly+').  BzrServerFactory can
        determine the original --directory from that transport.
        """
        # URLs always include the trailing slash, and get_base_path returns it
        base_dir = osutils.abspath("/a/b/c") + "/"
        base_url = urlutils.local_path_to_url(base_dir) + "/"
        # Define a fake 'protocol' to capture the transport that cmd_serve
        # passes to serve_bzr.

        def capture_transport(transport, host, port, inet, timeout):
            self.bzr_serve_transport = transport

        cmd = builtins.cmd_serve()
        # Read-only
        cmd.run(directory=base_dir, protocol=capture_transport)
        server_maker = BzrServerFactory()
        self.assertEqual("readonly+{}".format(base_url), self.bzr_serve_transport.base)
        self.assertEqual(base_dir, server_maker.get_base_path(self.bzr_serve_transport))
        # Read-write
        cmd.run(directory=base_dir, protocol=capture_transport, allow_writes=True)
        server_maker = BzrServerFactory()
        self.assertEqual(base_url, self.bzr_serve_transport.base)
        self.assertEqual(base_dir, server_maker.get_base_path(self.bzr_serve_transport))
        # Read-only, from a URL
        cmd.run(directory=base_url, protocol=capture_transport)
        server_maker = BzrServerFactory()
        self.assertEqual("readonly+{}".format(base_url), self.bzr_serve_transport.base)
        self.assertEqual(base_dir, server_maker.get_base_path(self.bzr_serve_transport))



# Copyright (C) 2006, 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""A generator which creates a yaml stanza of the current tree info."""

import yaml

from breezy import errors, hooks
from breezy.revision import NULL_REVISION
from breezy.version_info_formats import VersionInfoBuilder, create_date_str


class YamlVersionInfoBuilder(VersionInfoBuilder):
    """This writes a yaml stream out."""

    def generate(self, to_file):
        info = {}
        revision_id = self._get_revision_id()
        if revision_id != NULL_REVISION:
            info["revision-id"] = revision_id.decode("utf-8")
            rev = self._branch.repository.get_revision(revision_id)
            info["date"] = create_date_str(rev.timestamp, rev.timezone)
            try:
                revno = self._get_revno_str(revision_id)
            except errors.GhostRevisionsHaveNoRevno:
                revno = None
            for hook in YamlVersionInfoBuilder.hooks["revision"]:
                hook(rev, info)
        else:
            revno = "0"

        info["build-date"] = create_date_str()
        if revno is not None:
            info["revno"] = revno

        if self._branch.nick is not None:
            info["branch-nick"] = self._branch.nick

        if self._check or self._include_file_revs:
            self._extract_file_revisions()

        if self._check:
            if self._clean:
                info["clean"] = True
            else:
                info["clean"] = False

        if self._include_history:
            log = []
            for (
                revision_id,
                message,
                timestamp,
                timezone,
            ) in self._iter_revision_history():
                log.append(
                    {
                        "id": revision_id.decode("utf-8"),
                        "message": message,
                        "date": create_date_str(timestamp, timezone),
                    }
                )
            info["revisions"] = log

        if self._include_file_revs:
            files = []
            for path in sorted(self._file_revisions.keys()):
                files.append(
                    {
                        "path": path,
                        "revision": (
                            self._file_revisions[path].decode("utf-8")
                            if isinstance(self._file_revisions[path], bytes)
                            else self._file_revisions[path]
                        ),
                    }
                )
            info["file-revisions"] = files

        yaml.dump(info, to_file)


class YamlVersionInfoBuilderHooks(hooks.Hooks):
    """Hooks for yaml-formatted version-info output."""

    def __init__(self):
        super().__init__(
            "breezy.version_info_formats.format_yaml", "YamlVersionInfoBuilder.hooks"
        )
        self.add_hook(
            "revision",
            "Invoked when adding information about a revision to the"
            " YAML stanza that is printed. revision is called with a"
            " revision object and a YAML stanza.",
            (3, 3),
        )


YamlVersionInfoBuilder.hooks = YamlVersionInfoBuilderHooks()  # type: ignore



# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
# cython: language_level=3

"""Pyrex extensions for converting chunks to lines."""


cdef extern from "python-compat.h":
    pass

from cpython.bytes cimport (PyBytes_AS_STRING, PyBytes_CheckExact,
                            PyBytes_FromStringAndSize, PyBytes_GET_SIZE)
from cpython.list cimport PyList_Append
from libc.string cimport memchr


def chunks_to_lines(chunks):
    """Re-split chunks into simple lines.

    Each entry in the result should contain a single newline at the end. Except
    for the last entry which may not have a final newline. If chunks is already
    a simple list of lines, we return it directly.

    :param chunks: An list/tuple of strings. If chunks is already a list of
        lines, then we will return it as-is.
    :return: A list of strings.
    """
    cdef char *c_str
    cdef char *newline
    cdef char *c_last
    cdef Py_ssize_t the_len
    cdef int last_no_newline

    # Check to see if the chunks are already lines
    last_no_newline = 0
    for chunk in chunks:
        if last_no_newline:
            # We have a chunk which followed a chunk without a newline, so this
            # is not a simple list of lines.
            break
        # Switching from PyBytes_AsStringAndSize to PyBytes_CheckExact and
        # then the macros GET_SIZE and AS_STRING saved us 40us / 470us.
        # It seems PyBytes_AsStringAndSize can actually trigger a conversion,
        # which we don't want anyway.
        if not PyBytes_CheckExact(chunk):
            raise TypeError('chunk is not a string')
        the_len = PyBytes_GET_SIZE(chunk)
        if the_len == 0:
            # An empty string is never a valid line
            break
        c_str = PyBytes_AS_STRING(chunk)
        c_last = c_str + the_len - 1
        newline = <char *>memchr(c_str, c'\n', the_len)
        if newline != c_last:
            if newline == NULL:
                # Missing a newline. Only valid as the last line
                last_no_newline = 1
            else:
                # There is a newline in the middle, we must resplit
                break
    else:
        # Everything was already a list of lines
        return chunks

    # We know we need to create a new list of lines
    lines = []
    tail = None # Any remainder from the previous chunk
    for chunk in chunks:
        if tail is not None:
            chunk = tail + chunk
            tail = None
        if not PyBytes_CheckExact(chunk):
            raise TypeError('chunk is not a string')
        the_len = PyBytes_GET_SIZE(chunk)
        if the_len == 0:
            # An empty string is never a valid line, and we don't need to
            # append anything
            continue
        c_str = PyBytes_AS_STRING(chunk)
        c_last = c_str + the_len - 1
        newline = <char *>memchr(c_str, c'\n', the_len)
        if newline == c_last:
            # A simple line
            PyList_Append(lines, chunk)
        elif newline == NULL:
            # A chunk without a newline, if this is the last entry, then we
            # allow it
            tail = chunk
        else:
            # We have a newline in the middle, loop until we've consumed all
            # lines
            while newline != NULL:
                line = PyBytes_FromStringAndSize(c_str, newline - c_str + 1)
                PyList_Append(lines, line)
                c_str = newline + 1
                if c_str > c_last: # We are done
                    break
                the_len = c_last - c_str + 1
                newline = <char *>memchr(c_str, c'\n', the_len)
                if newline == NULL:
                    tail = PyBytes_FromStringAndSize(c_str, the_len)
                    break
    if tail is not None:
        PyList_Append(lines, tail)
    return lines



# Copyright (C) 2007-2012 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import sys

from .. import branch as _mod_branch
from .. import controldir, info, tests, workingtree
from .. import repository as _mod_repository
from ..bzr import branch as _mod_bzrbranch


class TestInfo(tests.TestCaseWithTransport):
    def test_describe_standalone_layout(self):
        tree = self.make_branch_and_tree("tree")
        self.assertEqual("Empty control directory", info.describe_layout())
        self.assertEqual(
            "Unshared repository with trees and colocated branches",
            info.describe_layout(tree.branch.repository, control=tree.controldir),
        )
        tree.branch.repository.set_make_working_trees(False)
        self.assertEqual(
            "Unshared repository with colocated branches",
            info.describe_layout(tree.branch.repository, control=tree.controldir),
        )
        self.assertEqual(
            "Standalone branch",
            info.describe_layout(
                tree.branch.repository, tree.branch, control=tree.controldir
            ),
        )
        self.assertEqual(
            "Standalone branchless tree",
            info.describe_layout(
                tree.branch.repository, None, tree, control=tree.controldir
            ),
        )
        self.assertEqual(
            "Standalone tree",
            info.describe_layout(
                tree.branch.repository, tree.branch, tree, control=tree.controldir
            ),
        )
        tree.branch.bind(tree.branch)
        self.assertEqual(
            "Bound branch",
            info.describe_layout(
                tree.branch.repository, tree.branch, control=tree.controldir
            ),
        )
        self.assertEqual(
            "Checkout",
            info.describe_layout(
                tree.branch.repository, tree.branch, tree, control=tree.controldir
            ),
        )
        checkout = tree.branch.create_checkout("checkout", lightweight=True)
        self.assertEqual(
            "Lightweight checkout",
            info.describe_layout(
                checkout.branch.repository,
                checkout.branch,
                checkout,
                control=tree.controldir,
            ),
        )

    def test_describe_repository_layout(self):
        repository = self.make_repository(".", shared=True)
        tree = controldir.ControlDir.create_branch_convenience(
            "tree", force_new_tree=True
        ).controldir.open_workingtree()
        self.assertEqual(
            "Shared repository with trees and colocated branches",
            info.describe_layout(tree.branch.repository, control=tree.controldir),
        )
        repository.set_make_working_trees(False)
        self.assertEqual(
            "Shared repository with colocated branches",
            info.describe_layout(tree.branch.repository, control=tree.controldir),
        )
        self.assertEqual(
            "Repository branch",
            info.describe_layout(
                tree.branch.repository, tree.branch, control=tree.controldir
            ),
        )
        self.assertEqual(
            "Repository branchless tree",
            info.describe_layout(
                tree.branch.repository, None, tree, control=tree.controldir
            ),
        )
        self.assertEqual(
            "Repository tree",
            info.describe_layout(
                tree.branch.repository, tree.branch, tree, control=tree.controldir
            ),
        )
        tree.branch.bind(tree.branch)
        self.assertEqual(
            "Repository checkout",
            info.describe_layout(
                tree.branch.repository, tree.branch, tree, control=tree.controldir
            ),
        )
        checkout = tree.branch.create_checkout("checkout", lightweight=True)
        self.assertEqual(
            "Lightweight checkout",
            info.describe_layout(
                checkout.branch.repository,
                checkout.branch,
                checkout,
                control=tree.controldir,
            ),
        )

    def assertTreeDescription(self, format):
        """Assert a tree's format description matches expectations."""
        self.make_branch_and_tree("{}_tree".format(format), format=format)
        tree = workingtree.WorkingTree.open("{}_tree".format(format))
        self.assertEqual(
            format,
            info.describe_format(
                tree.controldir, tree.branch.repository, tree.branch, tree
            ),
        )

    def assertCheckoutDescription(self, format, expected=None):
        """Assert a checkout's format description matches expectations."""
        if expected is None:
            expected = format
        branch = self.make_branch("{}_cobranch".format(format), format=format)
        # this ought to be easier...
        branch.create_checkout(
            "{}_co".format(format), lightweight=True
        ).controldir.destroy_workingtree()
        control = controldir.ControlDir.open("{}_co".format(format))
        old_format = control._format.workingtree_format
        try:
            control._format.workingtree_format = (
                controldir.format_registry.make_controldir(format).workingtree_format
            )
            control.create_workingtree()
            tree = workingtree.WorkingTree.open("{}_co".format(format))
            format_description = info.describe_format(
                tree.controldir, tree.branch.repository, tree.branch, tree
            )
            self.assertEqual(
                expected,
                format_description,
                "checkout of format called {!r} was described as {!r}".format(
                    expected, format_description
                ),
            )
        finally:
            control._format.workingtree_format = old_format

    def assertBranchDescription(self, format, expected=None):
        """Assert branch's format description matches expectations."""
        if expected is None:
            expected = format
        self.make_branch("{}_branch".format(format), format=format)
        branch = _mod_branch.Branch.open("{}_branch".format(format))
        self.assertEqual(
            expected,
            info.describe_format(branch.controldir, branch.repository, branch, None),
        )

    def assertRepoDescription(self, format, expected=None):
        """Assert repository's format description matches expectations."""
        if expected is None:
            expected = format
        self.make_repository("{}_repo".format(format), format=format)
        repo = _mod_repository.Repository.open("{}_repo".format(format))
        self.assertEqual(
            expected, info.describe_format(repo.controldir, repo, None, None)
        )

    def test_describe_tree_format(self):
        for key, format in controldir.format_registry.iteritems():
            if key in controldir.format_registry.aliases():
                continue
            if not format().supports_workingtrees:
                continue
            self.assertTreeDescription(key)

    def test_describe_checkout_format(self):
        for key in controldir.format_registry.keys():
            if key in controldir.format_registry.aliases():
                # Aliases will not describe correctly in the UI because the
                # real format is found.
                continue
            # legacy: weave does not support checkouts
            if key == "weave":
                continue
            # foreign: git checkouts can actually be bzr controldirs
            if key in ("git", "git-bare"):
                continue
            if controldir.format_registry.get_info(key).experimental:
                # We don't require that experimental formats support checkouts
                # or describe correctly in the UI.
                continue
            if controldir.format_registry.get_info(key).hidden:
                continue
            expected = None
            if key in ("pack-0.92",):
                expected = "pack-0.92"
            elif key in ("knit", "metaweave"):
                if "metaweave" in controldir.format_registry:
                    expected = "knit or metaweave"
                else:
                    expected = "knit"
            elif key in ("1.14", "1.14-rich-root"):
                expected = "1.14 or 1.14-rich-root"
            self.assertCheckoutDescription(key, expected)

    def test_describe_branch_format(self):
        for key in controldir.format_registry.keys():
            if key in controldir.format_registry.aliases():
                continue
            if controldir.format_registry.get_info(key).hidden:
                continue
            expected = None
            if key in ("dirstate", "knit"):
                expected = "dirstate or knit"
            elif key in ("1.14",):
                expected = "1.14"
            elif key in ("1.14-rich-root",):
                expected = "1.14-rich-root"
            self.assertBranchDescription(key, expected)

    def test_describe_repo_format(self):
        for key in controldir.format_registry.keys():
            if key in controldir.format_registry.aliases():
                continue
            if controldir.format_registry.get_info(key).hidden:
                continue
            expected = None
            if key in ("dirstate", "knit", "dirstate-tags"):
                expected = "dirstate or dirstate-tags or knit"
            elif key in ("1.14",):
                expected = "1.14"
            elif key in ("1.14-rich-root",):
                expected = "1.14-rich-root"
            self.assertRepoDescription(key, expected)

        format = controldir.format_registry.make_controldir("knit")
        format.set_branch_format(_mod_bzrbranch.BzrBranchFormat6())
        tree = self.make_branch_and_tree("unknown", format=format)
        self.assertEqual(
            "unnamed",
            info.describe_format(
                tree.controldir, tree.branch.repository, tree.branch, tree
            ),
        )

    def test_gather_location_controldir_only(self):
        bzrdir = self.make_controldir(".")
        self.assertEqual(
            [("control directory", bzrdir.user_url)],
            info.gather_location_info(control=bzrdir),
        )

    def test_gather_location_standalone(self):
        tree = self.make_branch_and_tree("tree")
        self.assertEqual(
            [("branch root", tree.controldir.root_transport.base)],
            info.gather_location_info(
                tree.branch.repository, tree.branch, tree, control=tree.controldir
            ),
        )
        self.assertEqual(
            [("branch root", tree.controldir.root_transport.base)],
            info.gather_location_info(
                tree.branch.repository, tree.branch, control=tree.controldir
            ),
        )
        return tree

    def test_gather_location_repo(self):
        srepo = self.make_repository("shared", shared=True)
        self.assertEqual(
            [("shared repository", srepo.controldir.root_transport.base)],
            info.gather_location_info(srepo, control=srepo.controldir),
        )
        urepo = self.make_repository("unshared")
        self.assertEqual(
            [("repository", urepo.controldir.root_transport.base)],
            info.gather_location_info(urepo, control=urepo.controldir),
        )

    def test_gather_location_repo_branch(self):
        srepo = self.make_repository("shared", shared=True)
        self.assertEqual(
            [("shared repository", srepo.controldir.root_transport.base)],
            info.gather_location_info(srepo, control=srepo.controldir),
        )
        tree = self.make_branch_and_tree("shared/tree")
        self.assertEqual(
            [
                ("shared repository", srepo.controldir.root_transport.base),
                ("repository branch", tree.branch.base),
            ],
            info.gather_location_info(srepo, tree.branch, tree, srepo.controldir),
        )

    def test_gather_location_light_checkout(self):
        tree = self.make_branch_and_tree("tree")
        lcheckout = tree.branch.create_checkout("lcheckout", lightweight=True)
        self.assertEqual(
            [
                ("light checkout root", lcheckout.controldir.root_transport.base),
                ("checkout of branch", tree.controldir.root_transport.base),
            ],
            self.gather_tree_location_info(lcheckout),
        )

    def test_gather_location_heavy_checkout(self):
        tree = self.make_branch_and_tree("tree")
        checkout = tree.branch.create_checkout("checkout")
        self.assertEqual(
            [
                ("checkout root", checkout.controldir.root_transport.base),
                ("checkout of branch", tree.controldir.root_transport.base),
            ],
            self.gather_tree_location_info(checkout),
        )
        light_checkout = checkout.branch.create_checkout(
            "light_checkout", lightweight=True
        )
        self.assertEqual(
            [
                ("light checkout root", light_checkout.controldir.root_transport.base),
                ("checkout root", checkout.controldir.root_transport.base),
                ("checkout of branch", tree.controldir.root_transport.base),
            ],
            self.gather_tree_location_info(light_checkout),
        )

    def test_gather_location_shared_repo_checkout(self):
        tree = self.make_branch_and_tree("tree")
        srepo = self.make_repository("shared", shared=True)
        shared_checkout = tree.branch.create_checkout("shared/checkout")
        self.assertEqual(
            [
                (
                    "repository checkout root",
                    shared_checkout.controldir.root_transport.base,
                ),
                ("checkout of branch", tree.controldir.root_transport.base),
                ("shared repository", srepo.controldir.root_transport.base),
            ],
            self.gather_tree_location_info(shared_checkout),
        )

    def gather_tree_location_info(self, tree):
        return info.gather_location_info(
            tree.branch.repository, tree.branch, tree, tree.controldir
        )

    def test_gather_location_bound(self):
        branch = self.make_branch("branch")
        bound_branch = self.make_branch("bound_branch")
        bound_branch.bind(branch)
        self.assertEqual(
            [
                ("branch root", bound_branch.controldir.root_transport.base),
                ("bound to branch", branch.controldir.root_transport.base),
            ],
            info.gather_location_info(
                bound_branch.repository, bound_branch, control=bound_branch.controldir
            ),
        )

    def test_gather_location_bound_in_repository(self):
        repo = self.make_repository("repo", shared=True)
        repo.set_make_working_trees(False)
        branch = self.make_branch("branch")
        bound_branch = controldir.ControlDir.create_branch_convenience(
            "repo/bound_branch"
        )
        try:
            bound_branch.bind(branch)
        except _mod_branch.BindingUnsupported:
            raise tests.TestNotApplicable("format does not support bound branches")
        self.assertEqual(
            [
                ("shared repository", bound_branch.repository.controldir.user_url),
                ("repository branch", bound_branch.controldir.user_url),
                ("bound to branch", branch.controldir.user_url),
            ],
            info.gather_location_info(bound_branch.repository, bound_branch),
        )

    def test_location_list(self):
        if sys.platform == "win32":
            raise tests.TestSkipped("Windows-unfriendly test")
        locs = info.LocationList("/home/foo")
        locs.add_url("a", "file:///home/foo/")
        locs.add_url("b", "file:///home/foo/bar/")
        locs.add_url("c", "file:///home/bar/bar")
        locs.add_url("d", "http://example.com/example/")
        locs.add_url("e", None)
        self.assertEqual(
            locs.locs,
            [
                ("a", "."),
                ("b", "bar"),
                ("c", "/home/bar/bar"),
                ("d", "http://example.com/example/"),
            ],
        )
        self.assertEqualDiff(
            "  a: .\n  b: bar\n  c: /home/bar/bar\n  d: http://example.com/example/\n",
            "".join(locs.get_lines()),
        )

    def test_gather_related_braches(self):
        branch = self.make_branch(".")
        branch.lock_write()
        try:
            branch.set_public_branch("baz")
            branch.set_push_location("bar")
            branch.set_parent("foo")
            branch.set_submit_branch("qux")
        finally:
            branch.unlock()
        self.assertEqual(
            [
                ("public branch", "baz"),
                ("push branch", "bar"),
                ("parent branch", "foo"),
                ("submit branch", "qux"),
            ],
            info._gather_related_branches(branch).locs,
        )



# Copyright (C) 2007-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for InventoryWorkingTree."""

from ..workingtree import InventoryModified
from . import TestCase


class ErrorTests(TestCase):
    def test_inventory_modified(self):
        error = InventoryModified("a tree to be repred")
        self.assertEqualDiff(
            "The current inventory for the tree 'a tree to "
            "be repred' has been modified, so a clean inventory cannot be "
            "read without data loss.",
            str(error),
        )



# Copyright (C) 2008, 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for check on a repository with external references."""

from breezy.tests.per_repository_reference import (
    TestCaseWithExternalReferenceRepository,
)


class TestCheck(TestCaseWithExternalReferenceRepository):
    def test_check_file_graph_across_external_boundary_ok(self):
        tree = self.make_branch_and_tree("base")
        self.build_tree(["base/file"])
        tree.add(["file"], ids=[b"file-id"])
        tree.commit("one")
        referring = self.make_branch_and_tree("referring")
        readonly_base = self.readonly_repository(tree.branch.repository)
        referring.branch.repository.add_fallback_repository(readonly_base)
        local_tree = referring.branch.create_checkout("local")
        self.build_tree_contents([("local/file", b"change")])
        local_tree.commit("two")
        check_result = referring.branch.repository.check(
            referring.branch.repository.all_revision_ids()
        )
        check_result.report_results(verbose=False)
        self.assertFalse("inconsistent parents" in self.get_log())



# Copyright (C) 2006-2009, 2011, 2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from breezy import conflicts, errors, osutils, revisiontree, tests
from breezy import transport as _mod_transport
from breezy.bzr import workingtree_4
from breezy.tests.per_tree import TestCaseWithTree
from breezy.tree import MissingNestedTree


class TestAnnotate(TestCaseWithTree):
    def test_annotate(self):
        work_tree = self.make_branch_and_tree("wt")
        tree = self.get_tree_no_parents_abc_content(work_tree)
        tree_revision = getattr(tree, "get_revision_id", lambda: b"current:")()
        tree.lock_read()
        self.addCleanup(tree.unlock)
        for revision, line in tree.annotate_iter("a"):
            self.assertEqual(b"contents of a\n", line)
            self.assertEqual(tree_revision, revision)
        tree_revision = getattr(tree, "get_revision_id", lambda: b"random:")()
        for revision, line in tree.annotate_iter("a", default_revision=b"random:"):
            self.assertEqual(b"contents of a\n", line)
            self.assertEqual(tree_revision, revision)


class TestPlanFileMerge(TestCaseWithTree):
    def test_plan_file_merge(self):
        work_a = self.make_branch_and_tree("wta")
        self.build_tree_contents([("wta/file", b"a\nb\nc\nd\n")])
        work_a.add("file")
        work_a.commit("base version")
        work_b = work_a.controldir.sprout("wtb").open_workingtree()
        self.build_tree_contents([("wta/file", b"b\nc\nd\ne\n")])
        tree_a = self.workingtree_to_test_tree(work_a)
        if getattr(tree_a, "plan_file_merge", None) is None:
            raise tests.TestNotApplicable("Tree does not support plan_file_merge")
        tree_a.lock_read()
        self.addCleanup(tree_a.unlock)
        self.build_tree_contents([("wtb/file", b"a\nc\nd\nf\n")])
        tree_b = self.workingtree_to_test_tree(work_b)
        tree_b.lock_read()
        self.addCleanup(tree_b.unlock)
        self.assertEqual(
            [
                ("killed-a", b"a\n"),
                ("killed-b", b"b\n"),
                ("unchanged", b"c\n"),
                ("unchanged", b"d\n"),
                ("new-a", b"e\n"),
                ("new-b", b"f\n"),
            ],
            list(tree_a.plan_file_merge("file", tree_b)),
        )


class TestReference(TestCaseWithTree):
    def skip_if_no_reference(self, tree):
        if not tree.supports_tree_reference():
            raise tests.TestNotApplicable("Tree references not supported")

    def create_nested(self):
        work_tree = self.make_branch_and_tree("wt")
        with work_tree.lock_write():
            self.skip_if_no_reference(work_tree)
            subtree = self.make_branch_and_tree("wt/subtree")
            subtree.commit("foo")
            work_tree.add_reference(subtree)
        tree = self._convert_tree(work_tree)
        self.skip_if_no_reference(tree)
        return tree, subtree

    def test_get_reference_revision(self):
        tree, subtree = self.create_nested()
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual(
            subtree.last_revision(), tree.get_reference_revision("subtree")
        )

    def test_iter_references(self):
        tree, subtree = self.create_nested()
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual(["subtree"], list(tree.iter_references()))

    def test_get_nested_tree(self):
        tree, subtree = self.create_nested()
        try:
            changes = subtree.changes_from(tree.get_nested_tree("subtree"))
            self.assertFalse(changes.has_changed())
        except MissingNestedTree:
            # Also okay.
            pass

    def test_get_root_id(self):
        # trees should return some kind of root id; it can be none
        tree = self.make_branch_and_tree("tree")
        if not tree.supports_file_ids:
            raise tests.TestNotApplicable("file ids not supported")
        root_id = tree.path2id("")
        if root_id is not None:
            self.assertIsInstance(root_id, bytes)

    def test_is_versioned(self):
        tree = self.make_branch_and_tree("tree")
        self.assertTrue(tree.is_versioned(""))
        self.assertFalse(tree.is_versioned("blah"))
        self.build_tree(["tree/dir/", "tree/dir/file"])
        self.assertFalse(tree.is_versioned("dir"))
        self.assertFalse(tree.is_versioned("dir/"))
        tree.add(["dir", "dir/file"])
        self.assertTrue(tree.is_versioned("dir"))
        self.assertTrue(tree.is_versioned("dir/"))


class TestFileIds(TestCaseWithTree):
    def setUp(self):
        super().setUp()
        if not self.workingtree_format.supports_setting_file_ids:
            raise tests.TestNotApplicable(
                "working tree does not support setting file ids"
            )

    def test_id2path(self):
        # translate from file-id back to path
        work_tree = self.make_branch_and_tree("wt")
        tree = self.get_tree_no_parents_abc_content(work_tree)
        a_id = tree.path2id("a")
        with tree.lock_read():
            self.assertEqual("a", tree.id2path(a_id))
            # other ids give an error- don't return None for this case
            self.assertRaises(errors.NoSuchId, tree.id2path, b"a")

    def test_all_file_ids(self):
        work_tree = self.make_branch_and_tree("wt")
        tree = self.get_tree_no_parents_abc_content(work_tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual(
            tree.all_file_ids(),
            {
                tree.path2id("a"),
                tree.path2id(""),
                tree.path2id("b"),
                tree.path2id("b/c"),
            },
        )


class TestStoredKind(TestCaseWithTree):
    def test_stored_kind(self):
        tree = self.make_branch_and_tree("tree")
        work_tree = self.make_branch_and_tree("wt")
        tree = self.get_tree_no_parents_abc_content(work_tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual("file", tree.stored_kind("a"))
        self.assertEqual("directory", tree.stored_kind("b"))


class TestFileContent(TestCaseWithTree):
    def test_get_file(self):
        work_tree = self.make_branch_and_tree("wt")
        tree = self.get_tree_no_parents_abc_content_2(work_tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        # Test lookup without path works
        file_without_path = tree.get_file("a")
        try:
            lines = file_without_path.readlines()
            self.assertEqual([b"foobar\n"], lines)
        finally:
            file_without_path.close()

    def test_get_file_context_manager(self):
        work_tree = self.make_branch_and_tree("wt")
        tree = self.get_tree_no_parents_abc_content_2(work_tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        with tree.get_file("a") as f:
            self.assertEqual(b"foobar\n", f.read())

    def test_get_file_text(self):
        work_tree = self.make_branch_and_tree("wt")
        tree = self.get_tree_no_parents_abc_content_2(work_tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        # test read by path
        self.assertEqual(b"foobar\n", tree.get_file_text("a"))

    def test_get_file_lines(self):
        work_tree = self.make_branch_and_tree("wt")
        tree = self.get_tree_no_parents_abc_content_2(work_tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        # test read by path
        self.assertEqual([b"foobar\n"], tree.get_file_lines("a"))

    def test_get_file_lines_multi_line_breaks(self):
        work_tree = self.make_branch_and_tree("wt")
        self.build_tree_contents([("wt/foobar", b"a\rb\nc\r\nd")])
        work_tree.add("foobar")
        tree = self._convert_tree(work_tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual([b"a\rb\n", b"c\r\n", b"d"], tree.get_file_lines("foobar"))


class TestExtractFilesBytes(TestCaseWithTree):
    def test_iter_files_bytes(self):
        work_tree = self.make_branch_and_tree("wt")
        self.build_tree_contents(
            [("wt/foo", b"foo"), ("wt/bar", b"bar"), ("wt/baz", b"baz")]
        )
        work_tree.add(["foo", "bar", "baz"])
        tree = self._convert_tree(work_tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        extracted = {
            i: b"".join(b)
            for i, b in tree.iter_files_bytes(
                [("foo", "id1"), ("bar", "id2"), ("baz", "id3")]
            )
        }
        self.assertEqual(b"foo", extracted["id1"])
        self.assertEqual(b"bar", extracted["id2"])
        self.assertEqual(b"baz", extracted["id3"])
        self.assertRaises(
            _mod_transport.NoSuchFile,
            lambda: list(tree.iter_files_bytes([("qux", "file1-notpresent")])),
        )


class TestConflicts(TestCaseWithTree):
    def test_conflicts(self):
        """Tree.conflicts() should return a ConflictList instance."""
        work_tree = self.make_branch_and_tree("wt")
        tree = self._convert_tree(work_tree)
        self.assertIsInstance(tree.conflicts(), conflicts.ConflictList)


class TestIterEntriesByDir(TestCaseWithTree):
    def test_iteration_order(self):
        work_tree = self.make_branch_and_tree(".")
        self.build_tree(["a/", "a/b/", "a/b/c", "a/d/", "a/d/e", "f/", "f/g"])
        work_tree.add(["a", "a/b", "a/b/c", "a/d", "a/d/e", "f", "f/g"])
        tree = self._convert_tree(work_tree)
        output_order = [p for p, e in tree.iter_entries_by_dir()]
        self.assertEqual(
            ["", "a", "f", "a/b", "a/d", "a/b/c", "a/d/e", "f/g"], output_order
        )


class TestIterChildEntries(TestCaseWithTree):
    def test_iteration_order(self):
        work_tree = self.make_branch_and_tree(".")
        self.build_tree(["a/", "a/b/", "a/b/c", "a/d/", "a/d/e", "f/", "f/g"])
        work_tree.add(["a", "a/b", "a/b/c", "a/d", "a/d/e", "f", "f/g"])
        tree = self._convert_tree(work_tree)
        output = [e.name for e in tree.iter_child_entries("")]
        self.assertEqual({"a", "f"}, set(output))
        output = [e.name for e in tree.iter_child_entries("a")]
        self.assertEqual({"b", "d"}, set(output))

    def test_does_not_exist(self):
        work_tree = self.make_branch_and_tree(".")
        self.build_tree(["a/"])
        work_tree.add(["a"])
        tree = self._convert_tree(work_tree)
        self.assertRaises(
            _mod_transport.NoSuchFile, lambda: list(tree.iter_child_entries("unknown"))
        )


class TestExtras(TestCaseWithTree):
    def test_extras(self):
        work_tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/file", "tree/versioned-file"])
        work_tree.add(["file", "versioned-file"])
        work_tree.commit("add files")
        work_tree.remove("file")
        tree = self._convert_tree(work_tree)
        if isinstance(
            tree, (revisiontree.RevisionTree, workingtree_4.DirStateRevisionTree)
        ):
            expected = []
        else:
            expected = ["file"]
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual(expected, list(tree.extras()))


class TestGetFileSha1(TestCaseWithTree):
    def test_get_file_sha1(self):
        work_tree = self.make_branch_and_tree("tree")
        self.build_tree_contents([("tree/file", b"file content")])
        work_tree.add("file")
        tree = self._convert_tree(work_tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        expected = osutils.sha_string(b"file content")
        self.assertEqual(expected, tree.get_file_sha1("file"))


class TestGetFileVerifier(TestCaseWithTree):
    def test_get_file_verifier(self):
        work_tree = self.make_branch_and_tree("tree")
        self.build_tree_contents(
            [("tree/file1", b"file content"), ("tree/file2", b"file content")]
        )
        work_tree.add(["file1", "file2"])
        tree = self._convert_tree(work_tree)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        (kind, data) = tree.get_file_verifier("file1")
        self.assertEqual(
            tree.get_file_verifier("file1"), tree.get_file_verifier("file2")
        )
        if kind == "SHA1":
            expected = osutils.sha_string(b"file content")
            self.assertEqual(expected, data)


class TestHasVersionedDirectories(TestCaseWithTree):
    def test_has_versioned_directories(self):
        work_tree = self.make_branch_and_tree("tree")
        tree = self._convert_tree(work_tree)
        self.assertIn(tree.has_versioned_directories(), (True, False))


class TestSupportsRenameTracking(TestCaseWithTree):
    def test_supports_rename_tracking(self):
        work_tree = self.make_branch_and_tree("tree")
        tree = self._convert_tree(work_tree)
        self.assertSubset([tree.supports_rename_tracking()], (True, False))


class TestSupportsVersionableKind(TestCaseWithTree):
    def test_file(self):
        work_tree = self.make_branch_and_tree("tree")
        tree = self._convert_tree(work_tree)
        self.assertTrue(tree.versionable_kind("file"))

    def test_unknown(self):
        work_tree = self.make_branch_and_tree("tree")
        tree = self._convert_tree(work_tree)
        self.assertFalse(tree.versionable_kind("unknown"))


class TestSpecialFilename(TestCaseWithTree):
    def test_is_special_path(self):
        work_tree = self.make_branch_and_tree("tree")
        tree = self._convert_tree(work_tree)
        self.assertFalse(tree.is_special_path("foo"))



#! /usr/bin/env python

# Copyright (C) 2005 Canonical Ltd

"""Print to stdout a description of the current directory,
formatted as a Python data structure.

This can be useful in tests that need to recreate directory
contents.
"""

import sys

from breezy.trace import enable_default_logging

enable_default_logging()
from breezy.selftest.treeshape import capture_tree_contents


def main(argv):
    # a lame reimplementation of pformat that splits multi-line
    # strings into concatenated string literals.
    print("[")
    for tt in capture_tree_contents("."):
        assert isinstance(tt, tuple)
        print("    (", repr(tt[0]) + ",", end=" ")
        if len(tt) == 1:
            print("),")
        else:
            assert len(tt) == 2
            val = tt[1]
            print()
            if val == "":
                print("        ''")
            else:
                for valline in val.splitlines(True):
                    print("       ", repr(valline))
            print("    ),")
    print("]")


if __name__ == "__main__":
    sys.exit(main(sys.argv))



# Copyright (C) 2010-2018 Jelmer Vernooij <jelmer@jelmer.uk>
# Copyright (C) 2011 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for Git working trees."""

import os
import stat

from dulwich import __version__ as dulwich_version
from dulwich.diff_tree import RenameDetector, tree_changes
from dulwich.index import ConflictedIndexEntry, IndexEntry
from dulwich.object_store import OverlayObjectStore
from dulwich.objects import S_IFGITLINK, ZERO_SHA, Blob, Tree

from ... import conflicts as _mod_conflicts
from ... import workingtree as _mod_workingtree
from ...bzr.inventorytree import InventoryTreeChange as TreeChange
from ...delta import TreeDelta
from ...tests import TestCase, TestCaseWithTransport
from ..mapping import default_mapping
from ..tree import tree_delta_from_git_changes


def changes_between_git_tree_and_working_copy(
    source_store,
    from_tree_sha,
    target,
    want_unchanged=False,
    want_unversioned=False,
    rename_detector=None,
    include_trees=True,
):
    """Determine the changes between a git tree and a working tree with index."""
    to_tree_sha, extras = target.git_snapshot(want_unversioned=want_unversioned)
    store = OverlayObjectStore([source_store, target.store])
    return tree_changes(
        store,
        from_tree_sha,
        to_tree_sha,
        include_trees=include_trees,
        rename_detector=rename_detector,
        want_unchanged=want_unchanged,
        change_type_same=True,
    ), extras


class GitWorkingTreeTests(TestCaseWithTransport):
    def setUp(self):
        super().setUp()
        self.tree = self.make_branch_and_tree(".", format="git")

    def test_conflict_list(self):
        self.assertIsInstance(self.tree.conflicts(), _mod_conflicts.ConflictList)

    def test_add_conflict(self):
        self.build_tree(["conflicted"])
        self.tree.add(["conflicted"])
        with self.tree.lock_tree_write():
            self.tree.index[b"conflicted"] = ConflictedIndexEntry(
                this=self.tree.index[b"conflicted"]
            )
            self.tree._index_dirty = True
        conflicts = self.tree.conflicts()
        self.assertEqual(1, len(conflicts))

    def test_revert_empty(self):
        self.build_tree(["a"])
        self.tree.add(["a"])
        self.assertTrue(self.tree.is_versioned("a"))
        self.tree.revert(["a"])
        self.assertFalse(self.tree.is_versioned("a"))

    def test_is_ignored_directory(self):
        self.assertFalse(self.tree.is_ignored("a"))
        self.build_tree(["a/"])
        self.assertFalse(self.tree.is_ignored("a"))
        self.build_tree_contents([(".gitignore", "a\n")])
        self.tree._ignoremanager = None
        self.assertTrue(self.tree.is_ignored("a"))
        self.build_tree_contents([(".gitignore", "a/\n")])
        self.tree._ignoremanager = None
        self.assertTrue(self.tree.is_ignored("a"))

    def test_add_submodule_dir(self):
        subtree = self.make_branch_and_tree("asub", format="git")
        subtree.commit("Empty commit")
        self.tree.add(["asub"])
        with self.tree.lock_read():
            entry = self.tree.index[b"asub"]
            self.assertEqual(entry.mode, S_IFGITLINK)
        self.assertEqual([], list(subtree.unknowns()))

    def test_add_submodule_file(self):
        os.mkdir(".git/modules")
        self.make_branch(".git/modules/asub", format="git-bare")
        os.mkdir("asub")
        with open("asub/.git", "w") as f:
            f.write("gitdir: ../.git/modules/asub\n")
        subtree = _mod_workingtree.WorkingTree.open("asub")
        subtree.commit("Empty commit")
        self.tree.add(["asub"])
        with self.tree.lock_read():
            entry = self.tree.index[b"asub"]
            self.assertEqual(entry.mode, S_IFGITLINK)
        self.assertEqual([], list(subtree.unknowns()))


class GitWorkingTreeFileTests(TestCaseWithTransport):
    def setUp(self):
        super().setUp()
        self.tree = self.make_branch_and_tree("actual", format="git")
        self.build_tree_contents(
            [("linked/",), ("linked/.git", "gitdir: ../actual/.git")]
        )
        self.wt = _mod_workingtree.WorkingTree.open("linked")

    def test_add(self):
        self.build_tree(["linked/somefile"])
        self.wt.add(["somefile"])
        self.wt.commit("Add somefile")


class TreeDeltaFromGitChangesTests(TestCase):
    def test_empty(self):
        delta = TreeDelta()
        changes = []
        self.assertEqual(
            delta,
            tree_delta_from_git_changes(changes, (default_mapping, default_mapping)),
        )

    def test_missing(self):
        delta = TreeDelta()
        delta.removed.append(
            TreeChange(
                b"git:a",
                ("a", "a"),
                False,
                (True, True),
                (b"TREE_ROOT", b"TREE_ROOT"),
                ("a", "a"),
                ("file", None),
                (True, False),
            )
        )
        changes = [
            ("remove", (b"a", stat.S_IFREG | 0o755, b"a" * 40), (b"a", 0, b"a" * 40))
        ]
        self.assertEqual(
            delta,
            tree_delta_from_git_changes(changes, (default_mapping, default_mapping)),
        )


class ChangesBetweenGitTreeAndWorkingCopyTests(TestCaseWithTransport):
    def setUp(self):
        super().setUp()
        self.wt = self.make_branch_and_tree(".", format="git")
        self.store = self.wt.branch.repository._git.object_store

    def expectDelta(
        self,
        expected_changes,
        expected_extras=None,
        want_unversioned=False,
        tree_id=None,
        rename_detector=None,
    ):
        if tree_id is None:
            try:
                tree_id = self.store[self.wt.branch.repository._git.head()].tree
            except KeyError:
                tree_id = None
        with self.wt.lock_read():
            changes, extras = changes_between_git_tree_and_working_copy(
                self.store,
                tree_id,
                self.wt,
                want_unversioned=want_unversioned,
                rename_detector=rename_detector,
            )
            self.assertEqual(expected_changes, list(changes))
        if expected_extras is None:
            expected_extras = set()
        self.assertEqual(set(expected_extras), set(extras))

    def test_empty(self):
        self.expectDelta([("add", (None, None, None), (b"", stat.S_IFDIR, Tree().id))])

    def test_added_file(self):
        self.build_tree(["a"])
        self.wt.add(["a"])
        a = Blob.from_string(b"contents of a\n")
        t = Tree()
        t.add(b"a", stat.S_IFREG | 0o644, a.id)
        self.expectDelta(
            [
                ("add", (None, None, None), (b"", stat.S_IFDIR, t.id)),
                ("add", (None, None, None), (b"a", stat.S_IFREG | 0o644, a.id)),
            ]
        )

    def test_renamed_file(self):
        self.build_tree(["a"])
        self.wt.add(["a"])
        self.wt.rename_one("a", "b")
        a = Blob.from_string(b"contents of a\n")
        self.store.add_object(a)
        oldt = Tree()
        oldt.add(b"a", stat.S_IFREG | 0o644, a.id)
        self.store.add_object(oldt)
        newt = Tree()
        newt.add(b"b", stat.S_IFREG | 0o644, a.id)
        self.store.add_object(newt)
        self.expectDelta(
            [
                ("modify", (b"", stat.S_IFDIR, oldt.id), (b"", stat.S_IFDIR, newt.id)),
                ("delete", (b"a", stat.S_IFREG | 0o644, a.id), (None, None, None)),
                ("add", (None, None, None), (b"b", stat.S_IFREG | 0o644, a.id)),
            ],
            tree_id=oldt.id,
        )
        if dulwich_version >= (0, 19, 15):
            self.expectDelta(
                [
                    (
                        "modify",
                        (b"", stat.S_IFDIR, oldt.id),
                        (b"", stat.S_IFDIR, newt.id),
                    ),
                    (
                        "rename",
                        (b"a", stat.S_IFREG | 0o644, a.id),
                        (b"b", stat.S_IFREG | 0o644, a.id),
                    ),
                ],
                tree_id=oldt.id,
                rename_detector=RenameDetector(self.store),
            )

    def test_copied_file(self):
        self.build_tree(["a"])
        self.wt.add(["a"])
        self.wt.copy_one("a", "b")
        a = Blob.from_string(b"contents of a\n")
        self.store.add_object(a)
        oldt = Tree()
        oldt.add(b"a", stat.S_IFREG | 0o644, a.id)
        self.store.add_object(oldt)
        newt = Tree()
        newt.add(b"a", stat.S_IFREG | 0o644, a.id)
        newt.add(b"b", stat.S_IFREG | 0o644, a.id)
        self.store.add_object(newt)
        self.expectDelta(
            [
                ("modify", (b"", stat.S_IFDIR, oldt.id), (b"", stat.S_IFDIR, newt.id)),
                ("add", (None, None, None), (b"b", stat.S_IFREG | 0o644, a.id)),
            ],
            tree_id=oldt.id,
        )

        if dulwich_version >= (0, 19, 15):
            self.expectDelta(
                [
                    (
                        "modify",
                        (b"", stat.S_IFDIR, oldt.id),
                        (b"", stat.S_IFDIR, newt.id),
                    ),
                    (
                        "copy",
                        (b"a", stat.S_IFREG | 0o644, a.id),
                        (b"b", stat.S_IFREG | 0o644, a.id),
                    ),
                ],
                tree_id=oldt.id,
                rename_detector=RenameDetector(self.store, find_copies_harder=True),
            )
            self.expectDelta(
                [
                    (
                        "modify",
                        (b"", stat.S_IFDIR, oldt.id),
                        (b"", stat.S_IFDIR, newt.id),
                    ),
                    ("add", (None, None, None), (b"b", stat.S_IFREG | 0o644, a.id)),
                ],
                tree_id=oldt.id,
                rename_detector=RenameDetector(self.store, find_copies_harder=False),
            )

    def test_added_unknown_file(self):
        self.build_tree(["a"])
        t = Tree()
        self.expectDelta([("add", (None, None, None), (b"", stat.S_IFDIR, t.id))])
        a = Blob.from_string(b"contents of a\n")
        t = Tree()
        t.add(b"a", stat.S_IFREG | 0o644, a.id)
        self.expectDelta(
            [
                ("add", (None, None, None), (b"", stat.S_IFDIR, t.id)),
                ("add", (None, None, None), (b"a", stat.S_IFREG | 0o644, a.id)),
            ],
            [b"a"],
            want_unversioned=True,
        )

    def test_missing_added_file(self):
        self.build_tree(["a"])
        self.wt.add(["a"])
        os.unlink("a")
        Blob.from_string(b"contents of a\n")
        t = Tree()
        t.add(b"a", 0, ZERO_SHA)
        self.expectDelta(
            [
                ("add", (None, None, None), (b"", stat.S_IFDIR, t.id)),
                ("add", (None, None, None), (b"a", 0, ZERO_SHA)),
            ],
            [],
        )

    def test_missing_versioned_file(self):
        self.build_tree(["a"])
        self.wt.add(["a"])
        self.wt.commit("")
        os.unlink("a")
        a = Blob.from_string(b"contents of a\n")
        oldt = Tree()
        oldt.add(b"a", stat.S_IFREG | 0o644, a.id)
        newt = Tree()
        newt.add(b"a", 0, ZERO_SHA)
        self.expectDelta(
            [
                ("modify", (b"", stat.S_IFDIR, oldt.id), (b"", stat.S_IFDIR, newt.id)),
                ("modify", (b"a", stat.S_IFREG | 0o644, a.id), (b"a", 0, ZERO_SHA)),
            ]
        )

    def test_versioned_replace_by_dir(self):
        self.build_tree(["a"])
        self.wt.add(["a"])
        self.wt.commit("")
        os.unlink("a")
        os.mkdir("a")
        olda = Blob.from_string(b"contents of a\n")
        oldt = Tree()
        oldt.add(b"a", stat.S_IFREG | 0o644, olda.id)
        newt = Tree()
        newa = Tree()
        newt.add(b"a", stat.S_IFDIR, newa.id)
        self.expectDelta(
            [
                ("modify", (b"", stat.S_IFDIR, oldt.id), (b"", stat.S_IFDIR, newt.id)),
                (
                    "modify",
                    (b"a", stat.S_IFREG | 0o644, olda.id),
                    (b"a", stat.S_IFDIR, newa.id),
                ),
            ],
            want_unversioned=False,
        )
        self.expectDelta(
            [
                ("modify", (b"", stat.S_IFDIR, oldt.id), (b"", stat.S_IFDIR, newt.id)),
                (
                    "modify",
                    (b"a", stat.S_IFREG | 0o644, olda.id),
                    (b"a", stat.S_IFDIR, newa.id),
                ),
            ],
            want_unversioned=True,
        )

    def test_extra(self):
        self.build_tree(["a"])
        newa = Blob.from_string(b"contents of a\n")
        newt = Tree()
        newt.add(b"a", stat.S_IFREG | 0o644, newa.id)
        self.expectDelta(
            [
                ("add", (None, None, None), (b"", stat.S_IFDIR, newt.id)),
                ("add", (None, None, None), (b"a", stat.S_IFREG | 0o644, newa.id)),
            ],
            [b"a"],
            want_unversioned=True,
        )

    def test_submodule(self):
        self.subtree = self.make_branch_and_tree("a", format="git")
        a = Blob.from_string(b"irrelevant\n")
        self.build_tree_contents([("a/.git/HEAD", a.id)])
        with self.wt.lock_tree_write():
            (index, index_path) = self.wt._lookup_index(b"a")
            index[b"a"] = IndexEntry(0, 0, 0, 0, S_IFGITLINK, 0, 0, 0, a.id)
            self.wt._index_dirty = True
        t = Tree()
        t.add(b"a", S_IFGITLINK, a.id)
        self.store.add_object(t)
        self.expectDelta([], tree_id=t.id)

    def test_submodule_not_checked_out(self):
        a = Blob.from_string(b"irrelevant\n")
        with self.wt.lock_tree_write():
            (index, index_path) = self.wt._lookup_index(b"a")
            index[b"a"] = IndexEntry(0, 0, 0, 0, S_IFGITLINK, 0, 0, 0, a.id)
            self.wt._index_dirty = True
        os.mkdir(self.wt.abspath("a"))
        t = Tree()
        t.add(b"a", S_IFGITLINK, a.id)
        self.store.add_object(t)
        self.expectDelta([], tree_id=t.id)



# Copyright (C) 2020 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for breezy.branch.InterBranch.copy_content_into."""

from breezy.tests.per_intertree import TestCaseWithTwoTrees
from breezy.transport import NoSuchFile


class TestFindPaths(TestCaseWithTwoTrees):
    def test_path_missing(self):
        tree1 = self.make_branch_and_tree("1")
        tree2 = self.make_to_branch_and_tree("2")
        self.build_tree_contents(
            [
                ("1/file", b"apples"),
                ("2/file", b"apples"),
            ]
        )
        tree1.add("file")
        tree2.add("file")
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        inter = self.intertree_class(tree1, tree2)
        self.assertRaises(NoSuchFile, inter.find_source_path, "missing")
        self.assertRaises(NoSuchFile, inter.find_target_path, "missing")
        self.assertRaises(NoSuchFile, inter.find_source_paths, ["missing"])
        self.assertRaises(NoSuchFile, inter.find_target_paths, ["missing"])

    def test_old_path(self):
        tree1 = self.make_branch_and_tree("1")
        tree2 = self.make_to_branch_and_tree("2")
        self.build_tree_contents(
            [
                ("2/file", b"apples"),
            ]
        )
        tree2.add("file")
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        inter = self.intertree_class(tree1, tree2)
        self.assertIs(None, inter.find_source_path("file"))
        self.assertEqual({"file": None}, inter.find_source_paths(["file"]))

    def test_new_path(self):
        tree1 = self.make_branch_and_tree("1")
        tree2 = self.make_to_branch_and_tree("2")
        self.build_tree_contents(
            [
                ("1/file", b"apples"),
            ]
        )
        tree1.add("file")
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        inter = self.intertree_class(tree1, tree2)
        self.assertIs(None, inter.find_target_path("file"))
        self.assertEqual({"file": None}, inter.find_target_paths(["file"]))

    def test_unchanged(self):
        tree1 = self.make_branch_and_tree("1")
        self.build_tree_contents(
            [
                ("1/file", b"apples"),
            ]
        )
        tree1.add("file")
        tree1.commit("foo")

        tree2 = self.make_to_branch_and_tree("2")
        tree2.pull(tree1.branch)
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        inter = self.intertree_class(tree1, tree2)
        self.assertEqual("file", inter.find_target_path("file"))
        self.assertEqual({"file": "file"}, inter.find_target_paths(["file"]))
        self.assertEqual("file", inter.find_source_path("file"))
        self.assertEqual({"file": "file"}, inter.find_source_paths(["file"]))

    def test_rename(self):
        tree1 = self.make_branch_and_tree("1")
        self.build_tree_contents(
            [
                ("1/file", b"apples"),
            ]
        )
        tree1.add("file")
        tree1.commit("foo")

        tree2 = self.make_to_branch_and_tree("2")
        tree2.pull(tree1.branch)
        tree2.rename_one("file", "newfile")
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        inter = self.intertree_class(tree1, tree2)
        self.assertEqual("newfile", inter.find_target_path("file"))
        self.assertEqual({"file": "newfile"}, inter.find_target_paths(["file"]))
        self.assertEqual("file", inter.find_source_path("newfile"))
        self.assertEqual({"newfile": "file"}, inter.find_source_paths(["newfile"]))

    def test_copy(self):
        tree1 = self.make_branch_and_tree("1")
        self.build_tree_contents(
            [
                ("1/file", b"apples"),
            ]
        )
        tree1.add("file")
        tree1.commit("foo")

        tree2 = self.make_to_branch_and_tree("2")
        tree2.pull(tree1.branch)
        tree2.copy_one("file", "newfile")
        tree1, tree2 = self.mutable_trees_to_test_trees(self, tree1, tree2)
        inter = self.intertree_class(tree1, tree2)
        self.assertIn(inter.find_target_path("file"), ["file", "newfile"])
        self.assertIn(inter.find_source_path("newfile"), ["file", None])
        self.assertEqual("file", inter.find_source_path("file"))
        self.assertIn(
            inter.find_source_paths(["newfile"]),
            [{"newfile": "file"}, {"newfile": None}],
        )

        self.assertIn(
            inter.find_source_paths(["newfile", "file"]),
            [
                {"newfile": "file", "file": "file"},
                {"newfile": None, "file": "file"},
            ],
        )



# Copyright (C) 2005-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test commit message editor."""

import os
import sys

from .. import commit, config, msgeditor, osutils, trace
from .. import transport as _mod_transport
from ..msgeditor import (
    edit_commit_message_encoded,
    make_commit_message_template_encoded,
)
from ..trace import mutter
from . import (
    TestCaseInTempDir,
    TestCaseWithTransport,
    TestNotApplicable,
    features,
    multiply_tests,
    probe_bad_non_ascii,
    probe_unicode_in_user_encoding,
    split_suite_by_re,
)
from .EncodingAdapter import encoding_scenarios


def load_tests(loader, standard_tests, pattern):
    """Parameterize the test for tempfile creation with different encodings."""
    to_adapt, result = split_suite_by_re(
        standard_tests, "test__create_temp_file_with_commit_template_in_unicode_dir"
    )
    return multiply_tests(to_adapt, encoding_scenarios, result)


class MsgEditorTest(TestCaseWithTransport):
    def make_uncommitted_tree(self):
        """Build a branch with uncommitted unicode named changes in the cwd."""
        working_tree = self.make_branch_and_tree(".")
        filename = "hell\u00d8"
        try:
            self.build_tree_contents([(filename, b"contents of hello")])
        except UnicodeEncodeError:
            self.skipTest(
                "can't build unicode working tree in filesystem encoding {}".format(
                    sys.getfilesystemencoding()
                )
            )
        working_tree.add(filename)
        return working_tree

    def test_commit_template(self):
        """Test building a commit message template."""
        working_tree = self.make_uncommitted_tree()
        template = msgeditor.make_commit_message_template(working_tree, None)
        self.assertEqualDiff(
            template,
            """\
added:
  hell\u00d8
""",
        )

    def make_multiple_pending_tree(self):
        config.GlobalStack().set("email", "Bilbo Baggins <bb@hobbit.net>")
        tree = self.make_branch_and_tree("a")
        tree.commit("Initial checkin.", timestamp=1230912900, timezone=0)
        tree2 = tree.controldir.clone("b").open_workingtree()
        tree.commit("Minor tweak.", timestamp=1231977840, timezone=0)
        tree2.commit("Feature X work.", timestamp=1233186240, timezone=0)
        tree3 = tree2.controldir.clone("c").open_workingtree()
        tree2.commit("Feature X finished.", timestamp=1233187680, timezone=0)
        tree3.commit(
            "Feature Y, based on initial X work.", timestamp=1233285960, timezone=0
        )
        tree.merge_from_branch(tree2.branch)
        tree.merge_from_branch(tree3.branch, force=True)
        return tree

    def test_commit_template_pending_merges(self):
        """Test building a commit message template when there are pending
        merges.  The commit message should show all pending merge revisions,
        as does 'status -v', not only the merge tips.
        """
        working_tree = self.make_multiple_pending_tree()
        template = msgeditor.make_commit_message_template(working_tree, None)
        self.assertEqualDiff(
            template,
            """\
pending merges:
  Bilbo Baggins 2009-01-29 Feature X finished.
    Bilbo Baggins 2009-01-28 Feature X work.
  Bilbo Baggins 2009-01-30 Feature Y, based on initial X work.
""",
        )

    def test_commit_template_encoded(self):
        """Test building a commit message template."""
        working_tree = self.make_uncommitted_tree()
        template = make_commit_message_template_encoded(
            working_tree, None, output_encoding="utf8"
        )
        self.assertEqualDiff(
            template,
            """\
added:
  hell\u00d8
""".encode(),
        )

    def test_commit_template_and_diff(self):
        """Test building a commit message template."""
        working_tree = self.make_uncommitted_tree()
        template = make_commit_message_template_encoded(
            working_tree, None, diff=True, output_encoding="utf8"
        )

        self.assertTrue(
            b"""\
@@ -0,0 +1,1 @@
+contents of hello
"""
            in template
        )
        self.assertTrue(
            """\
added:
  hell\u00d8
""".encode()
            in template
        )

    def make_do_nothing_editor(self, basename="fed"):
        if sys.platform == "win32":
            name = basename + ".bat"
            with open(name, "w") as f:
                f.write("@rem dummy fed")
            return name
        else:
            name = basename + ".sh"
            with open(name, "wb") as f:
                f.write(b"#!/bin/sh\n")
            os.chmod(name, 0o755)
            return "./" + name

    def test_run_editor(self):
        self.overrideEnv("BRZ_EDITOR", self.make_do_nothing_editor())
        self.assertEqual(
            True, msgeditor._run_editor(""), "Unable to run dummy fake editor"
        )

    def test_parse_editor_name(self):
        """Correctly interpret names with spaces.

        See <https://bugs.launchpad.net/bzr/+bug/220331>
        """
        self.overrideEnv(
            "BRZ_EDITOR", '"{}"'.format(self.make_do_nothing_editor("name with spaces"))
        )
        self.assertEqual(True, msgeditor._run_editor("a_filename"))

    def make_fake_editor(self, message="test message from fed\n"):
        """Set up environment so that an editor will be a known script.

        Sets up BRZ_EDITOR so that if an editor is spawned it will run a
        script that just adds a known message to the start of the file.
        """
        if not isinstance(message, bytes):
            message = message.encode("utf-8")
        with open("fed.py", "w") as f:
            f.write("#!{}\n".format(sys.executable))
            f.write(
                """\
# coding=utf-8
import sys
if len(sys.argv) == 2:
    fn = sys.argv[1]
    with open(fn, 'rb') as f:
        s = f.read()
    with open(fn, 'wb') as f:
        f.write({!r})
        f.write(s)
""".format(message)
            )
        if sys.platform == "win32":
            # [win32] make batch file and set BRZ_EDITOR
            with open("fed.bat", "w") as f:
                f.write(
                    """\
@echo off
"{}" fed.py %1
""".format(sys.executable)
                )
            self.overrideEnv("BRZ_EDITOR", "fed.bat")
        else:
            # [non-win32] make python script executable and set BRZ_EDITOR
            os.chmod("fed.py", 0o755)
            mutter("Setting BRZ_EDITOR to %r", "{} ./fed.py".format(sys.executable))
            self.overrideEnv("BRZ_EDITOR", "{} ./fed.py".format(sys.executable))

    def test_edit_commit_message_without_infotext(self):
        self.make_uncommitted_tree()
        self.make_fake_editor()

        mutter("edit_commit_message without infotext")
        self.assertEqual("test message from fed\n", msgeditor.edit_commit_message(""))

    def test_edit_commit_message_with_ascii_infotext(self):
        self.make_uncommitted_tree()
        self.make_fake_editor()

        mutter("edit_commit_message with ascii string infotext")
        self.assertEqual(
            "test message from fed\n", msgeditor.edit_commit_message("spam")
        )

    def test_edit_commit_message_with_unicode_infotext(self):
        self.make_uncommitted_tree()
        self.make_fake_editor()

        mutter("edit_commit_message with unicode infotext")
        uni_val, ue_val = probe_unicode_in_user_encoding()
        if ue_val is None:
            self.skipTest(
                "Cannot find a unicode character that works in encoding {}".format(
                    osutils.get_user_encoding()
                )
            )

        self.assertEqual(
            "test message from fed\n", msgeditor.edit_commit_message(uni_val)
        )

        tmpl = edit_commit_message_encoded("\u1234".encode())
        self.assertEqual("test message from fed\n", tmpl)

    def test_start_message(self):
        self.make_uncommitted_tree()
        self.make_fake_editor()
        self.assertEqual(
            "test message from fed\nstart message\n",
            msgeditor.edit_commit_message("", start_message="start message\n"),
        )
        self.assertEqual(
            "test message from fed\n",
            msgeditor.edit_commit_message("", start_message=""),
        )

    def test_deleted_commit_message(self):
        self.make_uncommitted_tree()

        if sys.platform == "win32":
            editor = "cmd.exe /c del"
        else:
            editor = "rm"
        self.overrideEnv("BRZ_EDITOR", editor)

        self.assertRaises(
            (EnvironmentError, _mod_transport.NoSuchFile),
            msgeditor.edit_commit_message,
            "",
        )

    def test__get_editor(self):
        self.overrideEnv("BRZ_EDITOR", "bzr_editor")
        self.overrideEnv("VISUAL", "visual")
        self.overrideEnv("EDITOR", "editor")

        conf = config.GlobalStack()
        conf.store._load_from_string(b"[DEFAULT]\neditor = config_editor\n")
        conf.store.save()
        editors = list(msgeditor._get_editor())
        editors = [editor for (editor, cfg_src) in editors]

        self.assertEqual(
            ["bzr_editor", "config_editor", "visual", "editor"], editors[:4]
        )

        if sys.platform == "win32":
            self.assertEqual(["wordpad.exe", "notepad.exe"], editors[4:])
        else:
            self.assertEqual(
                ["/usr/bin/editor", "vi", "pico", "nano", "joe"], editors[4:]
            )

    def test__run_editor_EACCES(self):
        """If running a configured editor raises EACESS, the user is warned."""
        self.overrideEnv("BRZ_EDITOR", "eacces.py")
        with open("eacces.py", "wb") as f:
            f.write(b"# Not a real editor")
        # Make the fake editor unreadable (and unexecutable)
        os.chmod("eacces.py", 0)
        # Set $EDITOR so that _run_editor will terminate before trying real
        # editors.
        self.overrideEnv("EDITOR", self.make_do_nothing_editor())
        # Call _run_editor, capturing mutter.warning calls.
        warnings = []

        def warning(*args):
            if len(args) > 1:
                warnings.append(args[0] % args[1:])
            else:
                warnings.append(args[0])

        _warning = trace.warning
        trace.warning = warning
        try:
            msgeditor._run_editor("")
        finally:
            trace.warning = _warning
        self.assertStartsWith(warnings[0], 'Could not start editor "eacces.py"')

    def test__create_temp_file_with_commit_template(self):
        # check that commit template written properly
        # and has platform native line-endings (CRLF on win32)
        create_file = msgeditor._create_temp_file_with_commit_template
        msgfilename, hasinfo = create_file(b"infotext", "----", b"start message")
        self.assertNotEqual(None, msgfilename)
        self.assertTrue(hasinfo)
        expected = os.linesep.join(["start message", "", "", "----", "", "infotext"])
        self.assertFileEqual(expected, msgfilename)

    def test__create_temp_file_with_commit_template_in_unicode_dir(self):
        self.requireFeature(features.UnicodeFilenameFeature)
        if hasattr(self, "info"):
            tmpdir = self.info["directory"]
            os.mkdir(tmpdir)
            # Force the creation of temp file in a directory whose name
            # requires some encoding support
            msgeditor._create_temp_file_with_commit_template(b"infotext", tmpdir=tmpdir)
        else:
            raise TestNotApplicable("Test run elsewhere with non-ascii data.")

    def test__create_temp_file_with_empty_commit_template(self):
        # empty file
        create_file = msgeditor._create_temp_file_with_commit_template
        msgfilename, hasinfo = create_file("")
        self.assertNotEqual(None, msgfilename)
        self.assertFalse(hasinfo)
        self.assertFileEqual("", msgfilename)

    def test_unsupported_encoding_commit_message(self):
        self.overrideEnv("LANG", "C")
        # LANG env variable has no effect on Windows
        # but some characters anyway cannot be represented
        # in default user encoding
        char = probe_bad_non_ascii(osutils.get_user_encoding())
        if char is None:
            self.skipTest(
                "Cannot find suitable non-ascii character "
                "for user_encoding ({})".format(osutils.get_user_encoding())
            )

        self.make_fake_editor(message=char)

        self.make_uncommitted_tree()
        self.assertRaises(
            msgeditor.BadCommitMessageEncoding, msgeditor.edit_commit_message, ""
        )

    def test_set_commit_message_no_hooks(self):
        commit_obj = commit.Commit()
        self.assertIs(None, msgeditor.set_commit_message(commit_obj))

    def test_set_commit_message_hook(self):
        msgeditor.hooks.install_named_hook(
            "set_commit_message",
            lambda commit_obj, existing_message: "save me some typing\n",
            None,
        )
        commit_obj = commit.Commit()
        self.assertEqual(
            "save me some typing\n", msgeditor.set_commit_message(commit_obj)
        )

    def test_generate_commit_message_template_no_hooks(self):
        commit_obj = commit.Commit()
        self.assertIs(None, msgeditor.generate_commit_message_template(commit_obj))

    def test_generate_commit_message_template_hook(self):
        msgeditor.hooks.install_named_hook(
            "commit_message_template",
            lambda commit_obj, msg: "save me some typing\n",
            None,
        )
        commit_obj = commit.Commit()
        self.assertEqual(
            "save me some typing\n",
            msgeditor.generate_commit_message_template(commit_obj),
        )


# GZ 2009-11-17: This wants moving to osutils when the errno checking code is
class TestPlatformErrnoWorkarounds(TestCaseInTempDir):
    """Ensuring workarounds enshrined in code actually serve a purpose."""

    def test_subprocess_call_bad_file(self):
        if sys.platform != "win32":
            raise TestNotApplicable("Workarounds for windows only")
        import errno
        import subprocess

        ERROR_BAD_EXE_FORMAT = 193
        open("textfile.txt", "w").close()
        e = self.assertRaises(WindowsError, subprocess.call, "textfile.txt")
        self.assertEqual(e.errno, errno.ENOEXEC)
        self.assertEqual(e.winerror, ERROR_BAD_EXE_FORMAT)
